from cpython cimport bool

# TODO: possibly use a STL map and pointers
#   to get better performance for searches?

__all__ = ["Trie", "MixedCaseSensitivityTrie"]

def longest_nonoverlapping_matches(matches):
    out = []
    cdef Match m1, m2
    cdef bool ok
    matches = sorted(matches, key=len, reverse=True)
    for m1 in matches:
        ok = True
        for m2 in out:
            if m1.overlaps(m2):
                ok = False
        if ok:
            out.append(m1)
    return out

def restrict_by_boundary_characters(matches, str query, str bchars):
    out = []
    for m in matches:
        if not ((m.start == 0) or (query[m.start - 1] in bchars)):
            continue
        if not ((m.end == len(query)) or (query[m.end] in bchars)):
            continue
        out.append(m)
    return out

cdef class Match:
    cdef readonly:
        int start, end
        object key

    def __cinit__(self, start, end, key=None):
        self.start = start
        self.end = end
        self.key = key

    def __len__(self):
        return self.end - self.start

    cdef bool overlaps(self, Match o):
        return self.end > o.start and self.start < o.end

    def __repr__(self):
        return "<acfsa.Match @ {}-{}, key={}>".format(self.start, self.end, repr(self.key))

cdef class Node:
    cdef public:
        int depth
        object key
        bool terminal
        dict children
        Node fail

    def __cinit__(self, int depth=0):
        self.terminal = False
        self.children = {}
        self.depth = depth

    def search(self, Py_UCS4 c):
        if <int> c in self.children:
            return self.children[<int> c]
        elif self.depth == 0:
            return self
        else:
            return self.fail.search(c)

    def search_without_fail_transition(self, Py_UCS4 c):
        return self.children.get(<int> c, self)

def _add_fail_transitions(node):
    for c,child in node.children.items():
        child.fail = node.fail.search_without_fail_transition(c)

    for child in node.children.values():
        _add_fail_transitions(child)

cdef class Trie:
    cdef:
        Node root
        bool built, case_sensitive, allow_overlaps
        str boundary_characters

    def __init__(self,
            case_sensitive=True,
            allow_overlaps=True,
            boundary_characters=""):
        """
        Create a keyword Trie.

        Parameters
        ----------

        case_sensitive : bool, optional
        allow_overlaps : bool, optional
            Whether to allow overlapping matches or not.
            If not, the longest nonoverlapping matches are returned.
        boundary_characters : str, optional
            If provided, matches are required to be bounded on both sides
            by one of the given characters, or by the query string beginning
            or end.
        """
        self.root = Node()
        self.built = False
        self.case_sensitive = case_sensitive
        self.allow_overlaps = allow_overlaps
        self.boundary_characters = boundary_characters

    def add(self, str text, key=None):
        if len(text) == 0:
            raise Exception("Cannot add empty text to Trie.")
        if self.built:
            raise Exception("Cannot add a new search term to a built Trie.")
        if not self.case_sensitive:
            text = text.lower()
        cdef Node node = self.root
        cdef Py_UCS4 c
        while text:
            c = text[0]
            if not <int> c in node.children:
                node.children[<int> c] = Node(depth=node.depth+1)
            node = node.children[<int> c]
            text = text[1:]
        node.terminal = True
        node.key = key

    def build(self):
        self.root.fail = self.root
        for child in self.root.children.values():
            child.fail = self.root
        for child in self.root.children.values():
            _add_fail_transitions(child)
        self.built = True

    def search(self, str text):
        if not self.built:
            raise Exception("Must call Trie.build() before Trie.search()") 
        if not self.case_sensitive:
            text = text.lower()
        cdef Py_UCS4 c
        cdef Node node = self.root
        cdef Node n
        matches = []
        for i,c in enumerate(text):
            node = node.search(c)
            if node.terminal:
                n = node
                while n.terminal:
                    matches.append(Match(1 + i - n.depth, i + 1, node.key))
                    n = n.fail
        if self.boundary_characters:
            matches = restrict_by_boundary_characters(
                    matches, text, self.boundary_characters)
        if not self.allow_overlaps:
            matches = longest_nonoverlapping_matches(matches)
        return matches

class MixedCaseSensitivityTrie(object):
    def __init__(self, allow_overlaps=True, boundary_characters=""):
        self._cs = Trie(case_sensitive=True,
                boundary_characters=boundary_characters,
                allow_overlaps=True)
        self._ci = Trie(case_sensitive=False,
                boundary_characters=boundary_characters,
                allow_overlaps=True)
        self.allow_overlaps = allow_overlaps

    def build(self):
        self._cs.build()
        self._ci.build()

    def add(self, str text, bool case_sensitive=False, object key=None):
        trie = self._cs if case_sensitive else self._ci
        trie.add(text, key=key)

    def search(self, str text):
        matches = self._cs.search(text) + self._ci.search(text)
        if not self.allow_overlaps:
            matches = longest_nonoverlapping_matches(matches)
        return matches



"""
Ontology handling (focused on Open Biomedical Ontology [OBO] format).
"""

import collections
import functools

import networkx as nx
import pandas as pd
import numpy as np

import acfsa
from wrenlab.util import memoize, download

Term = collections.namedtuple("Term", [
    "id", "name", "synonym", "relations", "namespace"
])

class Ontology(object):
    """
    Attribute specifications (DataFrame objects):
        terms: Term ID, Name, Depth, Namespace
        synonyms: Term ID, Synonym
        relations: Agent, Target, Relation
    """
    def __init__(self, terms, synonyms, relations, prefix=None):
        self._terms = terms
        self._synonyms = synonyms
        self._relations = relations

        if isinstance(prefix, str):
            self._prefix = prefix
        else:
            self._prefix = collections.Counter([t.split(":")[0] for t in self.terms.index])\
                    .most_common(1)[0][0]

    def __repr__(self):
        return "<Ontology with %s terms, %s synonyms, and %s relations>" \
                % (len(self._terms), len(self._synonyms), len(self._relations))

    @property
    def terms(self):
        return self._terms.copy()

    @property
    def synonyms(self):
        return self._synonyms.copy()

    @property
    def relations(self):
        return self._relations.copy()

    def _resolve_id(self, x):
        if isinstance(x, str):
            if x.startswith(self._prefix):
                return x
            else:
                return self._name_to_id(x)
        elif isinstance(x, int):
            return "{}:{:07d}".format(self._prefix, x)
        else:
            msg = "{} could not be resolved to a term ID in ontology {}"\
                .format(repr(x), self._prefix)
            raise ValueError(msg)

    def _name_to_id(self, name):
        ix = self.terms["Name"].str.lower() == name.lower()
        if ix.sum() == 1:
            return self.terms.ix[ix,:].index[0]
        raise KeyError

    def descendants(self, term):
        id = self._resolve_id(term)
        return nx.ancestors(self.graph, id)

    @property
    def depths(self):
        g = self.graph
        depths = nx.single_source_shortest_path_length(g.reverse(), g.root)
        o = {}
        for t in g.nodes():
            depth = depths.get(t)
            if depth is not None:
                o[t] = depth
        return pd.Series(o)

    @property
    def graph(self):
        if not hasattr(self, "_graph"):
            self._graph = self.to_graph()
        return self._graph
    
    def to_graph(self, relations=None):
        """
        Convert this Ontology into a NetworkX graph for efficient traversal.

        Parameters
        ----------
        relations : list of str or None, optional
            The relation types to use as edges in the graph. (default: is_a)

        Returns
        -------
        A :class:`networkx.DiGraph`. The edges are directed in the 
            same direction as the relation. In the common case of 
            the `is_a` relation, this means from children to parents.
        """
        g = nx.DiGraph()
        R = self.relations
        if relations is not None:
            R = R.ix[R["Relation"].isin(relations),:]

        for id,name,ns in self.terms.to_records():
            g.add_node(id, name=name, namespace=ns)
        for id,synonym in self.synonyms.to_records():
            if not "synonyms" in g.node[id]:
                g.node[id]["synonyms"] = []
            g.node[id]["synonyms"].append(synonym)
        for _,agent,target,relation in R.to_records():
            g.add_edge(agent, target, type=relation)

        g.root = sorted([n for n in g.nodes() if len(nx.descendants(g,n)) == 0],
                key=lambda n: -len(nx.ancestors(g,n)))[0]
        depths = nx.single_source_shortest_path_length(g.reverse(), g.root)
        o = {}
        for t in g.nodes():
            depth = depths.get(t)
            if depth is not None:
                g.node[t]["depth"] = depth
        return g

    def _expand_synonym(self, text):
        o = [text]
        if not text.endswith("s"):
            o.append(text+"s")
        return o

    @property
    def trie(self):
        if not hasattr(self, "_trie"):
            o = acfsa.MixedCaseSensitivityTrie(allow_overlaps=False)
            for ix,t in self.terms.loc[:,["Name"]].to_records():
                for s in self._expand_synonym(t):
                    o.add(s, key=ix, case_sensitive=(len(s) < 5))
            for ix,t in self.synonyms.loc[:,["Term ID", "Synonym"]].to_records(index=False):
                for s in self._expand_synonym(t):
                    o.add(s, key=ix, case_sensitive=(len(s) < 5))
            o.build()
            self._trie = o
        return self._trie

    def match(self, text):
        if text is None:
            return None
        trie = self.trie
        ms = trie.search(text)
        if len(ms) == 1:
            k = ms[0].key
            if isinstance(k, np.float64) and np.isnan(k):
                return
            return k

    @property
    def ancestry_table(self):
        g = self.to_graph()
        rows = []
        for n in g.nodes():
            for ancestor in nx.descendants(g, n):
                rows.append((ancestor, n))
        return pd.DataFrame(rows, 
                columns=["Ancestor", "Descendant"])

    def annotation_matrix(self, mapping, recursive=False):
        """
        Create a DataFrame representing the mapping between a set
        of ontology terms and another set of objects (genes, samples, etc.).

        Parameters
        ----------
        mapping : :class:`pandas.DataFrame`
            A DataFrame representing the mapping between ontology terms 
            and genes/samples/whatever. Two columns are required:
            Column 1 - The sample/gene ID
            Column 2 - The ontology term ID
            Column 3 (optional) - An integer, float, or boolean column. If not 
                provided, an int8 :class:`pandas.DataFrame` will be returned
                wherein all matching sample/ontology pairs will
                be assigned integer value 1 and non-matching pairs will be
                assigned 0.
        recursive : bool, optional
            If True, the ontology tree will be traversed and all parent
            terms will be assigned the same value

        Returns
        -------
        :class:`pandas.DataFrame`
            Rows are sample/gene IDs, columns are ontology IDs, and
            values are ints/floats/booleans as described above.
        """
        msg = "Provided DataFrame must have either two or three columns."
        assert mapping.shape[1] in (2,3), msg

        A = mapping
        if A.shape[1] == 2:
            A["Value"] = 1
            dtype = np.uint8
        else:
            dtype = mapping.iloc[2].dtype
            msg = "Value column must be numeric or boolean, not object."
            assert dtype is not np.object, msg

        A.columns = [A.columns[0]] + ["Term ID", "Value"]

        if recursive:
            inferred = mapping.merge(self.ancestry_table,
                    left_on=A.columns[1],
                    right_on="Descendant")\
                            .ix[:,["Ancestor", A.columns[0], "Value"]]
            inferred.columns = ["Term ID", A.columns[0], "Value"]
            A = pd.concat([A, inferred], axis=0).drop_duplicates()
        return A.pivot(A.columns[0], "Term ID", "Value")\
                .fillna(0).astype(dtype)

def _obo_make_term(attrs):
    if ("id" in attrs) and ("name" in attrs):
        for key in Term._fields:
            attrs.setdefault(key, [])
        attrs["namespace"] = attrs["namespace"] or None
        return Term(**attrs)

def _obo_parse(handle):
    """
    A simple iterative reader for OBO (Open Biomedical Ontology) files.
    
    :param handle: File handle to the OBO file
    :type handle: A file handle in 'rt' mode

    Currently, only a subset of the OBO specification is supported. 
    Namely, only the following attribute of [Term] entries:
    - id
    - name
    - relations
    - synonym
    - namespace
    """
    is_term = False
    attrs = collections.defaultdict(list)
    for line in handle:
        line = line.strip()
        if line in ("[Term]", "[Typedef]", "[Instance]"):
            is_term = line == "[Term]"
            t = _obo_make_term(attrs)
            if t and is_term: 
                yield t
            attrs = collections.defaultdict(list)
        elif line:
            try:
                key, value = line.split(": ", 1)
            except ValueError:
                continue

            if key == "id":
                attrs["id"] = value
            elif key == "name":
                attrs["name"] = value
            elif key == "is_a":
                attrs["relations"].append(("is_a", value.split("!")[0].strip()))
            elif key == "relationship":
                value = value.split("!")[0].strip()
                rel, target, *rest = value.split(" ")
                attrs["relations"].append((rel,target))
            elif key == "namespace":
                attrs["namespace"] = value
            elif key == "synonym":
                value = value[1:]
                attrs["synonym"].append(value[:value.find("\"")])
    t = _obo_make_term(attrs)
    if t and is_term: 
        yield t

def _obo_parse_as_data_frame(handle):
    terms, synonyms, relations = [], [], []

    for t in _obo_parse(handle):
        terms.append((t.id, t.name, t.namespace))
        for s in t.synonym:
            synonyms.append((t.id, s))
        synonyms.append((t.id, t.name))
        for rel, target in t.relations:
            relations.append((t.id, target, rel))

    terms = pd.DataFrame.from_records(terms,
            columns=["Term ID", "Name", "Namespace"],
            index="Term ID")
    synonyms = pd.DataFrame.from_records(synonyms,
            columns=["Term ID", "Synonym"])\
                    .drop_duplicates()
    relations = pd.DataFrame.from_records(relations,
            columns=["Agent", "Target", "Relation"])
    return terms, synonyms, relations
        
def parse(handle, format="obo"):
    assert (format=="obo"), "Only OBO format currently supported."
    return Ontology(*_obo_parse_as_data_frame(handle))

@memoize
def _fetch(abbreviation):
    """
    Fetch an OBO-format ontology from BerkeleyBOP.
    """
    url = "http://www.berkeleybop.org/ontologies/%s.obo" \
            % abbreviation.lower()
    path = download(url)
    for encoding in ["utf-8", "iso-8859-1"]:
        try:
            with open(str(path), encoding=encoding) as handle:
                return parse(handle, format="obo")
        except UnicodeDecodeError:
            continue

@functools.lru_cache()
def fetch(abbreviation):
    return _fetch(abbreviation)



"""
Utility to download, parse, and memoize remote data tables in various formats.

The specification for each table is found in wrenlab/yml/rtable.yml.

(TODO: document the YAML format).
"""

import collections
import functools
import os.path
import zipfile

import pandas as pd
import yaml

from wrenlab.util.log import LOG
from wrenlab.util.memoize import memoize
from wrenlab.util.net import download

################
# YAML utilities
################

# make PyYAML read dicts into OrderedDict
_mapping_tag = yaml.resolver.BaseResolver.DEFAULT_MAPPING_TAG

def dict_representer(dumper, data):
    return dumper.represent_dict(data.iteritems())

def dict_constructor(loader, node):
    return collections.OrderedDict(loader.construct_pairs(node))

yaml.add_representer(collections.OrderedDict, dict_representer)
yaml.add_constructor(_mapping_tag, dict_constructor)

@functools.lru_cache()
def _read_bundled_yaml(name):
    path = os.path.join(os.path.dirname(__file__), 
            "..",
            "yml",
            "{}.yml".format(name))
    with open(path) as h:
        return yaml.load(h)

@memoize
def remote_table(key, **kwargs):
    LOG.debug("Initializing remote table: {}".format(key))

    schemata = _read_bundled_yaml("rtable")
    schema = schemata[key]
    url = schema["url"]
    if "parameters" in schema:
        for k in schema["parameters"]:
            assert k in kwargs
        url = url.format(**kwargs)
    path = download(url)

    kwargs = {
        "compression": "gzip" if url.endswith(".gz") else None,
        "encoding": "utf-8"
    }

    columns = schema.get("columns")
    if isinstance(columns, dict):
        kwargs["names"] = list(schema["columns"].keys())
        kwargs["dtype"] = dict([(k,eval(v[1])) for k,v in schema["columns"].items()])
        kwargs["usecols"] = [v[0] for v in schema["columns"].values()]
        assert list(sorted(kwargs["usecols"])) == kwargs["usecols"]
    elif isinstance(columns, list):
        kwargs["names"] = columns

    if "kwargs" in schema:
        assert isinstance(schema["kwargs"], dict)
        kwargs.update(schema["kwargs"])

    archive = schema.get("archive")
    if archive is None:
        df = pd.read_table(path, **kwargs)
    else:
        type = archive["type"]
        name = archive["name"]
        if type == "zip":
            with zipfile.ZipFile(path) as obj:
                with obj.open(name) as h:
                    df = pd.read_table(h, **kwargs)
        else:
            raise NotImplementedError

    index_columns = schema.get("index")
    if index_columns is not None:
        df.drop_duplicates(index_columns, inplace=True)
        df = df.set_index(index_columns)

    return df





import os.path
import sqlite3
import warnings

import numpy as np
import pandas as pd

import acfsa
import wrenlab.ontology
import wrenlab.ncbi.geo.metadb

from wrenlab.util import *

EXCLUDE_TISSUES = [ "ear", "hip", "sputum", "pith", "neck", "tail", "stem",
        "wing", "stoma", "serum", "node", "fin", "finger", "primary cell",
        "root", "adult", "arm", "blast cell", "primary cell" ] 

INCLUDED_TISSUES = [89, 149, 763, 142, 759, 269, 784, 141, 1253, 1101, 930,
        876, 232, 1487, 772, 452, 290, 414, 991, 887, 801, 1629, 123, 671, 2807, 1158, 562]

def excluded_tissues():
    BTO = wrenlab.ontology.fetch("BTO")
    return [int(x[4:]) for x in BTO.terms.index[BTO.terms.Name.isin(EXCLUDE_TISSUES)]]

@memoize
def disease_state():
    """
    """
    # FIXME: use concat title + ch instead of just ch
    PATTERNS = {
            "Cancer": "(cancer|carcinoma|neoplas|leukemia)"
    }
    mdb = wrenlab.ncbi.geo.metadb.connect()
    df = mdb.samples
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        ix = df.Characteristics.dropna().str.contains(PATTERNS["Cancer"]).astype(float)
    o = pd.DataFrame.from_dict({"Cancer": ix})
    return o.dropna()

@memoize
def tissue():
    """
    """
    mdb = wrenlab.ncbi.geo.metadb.connect()
        
    BTO = wrenlab.ontology.fetch("BTO")
    ix = BTO.synonyms["Term ID"]\
            .isin(["BTO:{:07d}".format(ix) for ix in INCLUDED_TISSUES])
    synonyms = BTO.synonyms.ix[ix,:]

    trie = acfsa.Trie(
            case_sensitive=False, 
            allow_overlaps=False, 
            boundary_characters=" \n.:?")
    for term_id, s in synonyms.to_records(index=False):
        trie.add(s, key=int(term_id[4:]))
    trie.build()

    records = mdb.samples.loc[:,["Title", "Characteristics"]].to_records()

    o = []
    for sample_id, channel, title, ch in records:
        text = "\n".join(filter(None, [title,ch])).strip()
        if len(text) == 0:
            continue
        ms = trie.search(text)
        if len(ms) != 1:
            continue
        o.append((sample_id, channel, ms[0].key))
    o = pd.DataFrame.from_records(o, 
            columns=["SampleID", "Channel", "TissueID"])\
                    .set_index(["SampleID", "Channel"])
    o.TissueID = o.TissueID.astype("category")
    return o

@memoize
def map_experiments_to_samples(method="largest", min_size=10):
    assert method in ("largest", "smallest")
    assert isinstance(min_size, int)

    mdb = wrenlab.ncbi.geo.metadb.connect()
    j = mdb.experiment_sample
    counts = j.ExperimentID.value_counts()
    counts = counts.ix[counts >= min_size]
    counts = dict(zip(counts.index, counts))
    j["Count"] = [counts.get(ix) for ix in j.ExperimentID]
    def fn(df):
        return df.sort_values("Count", ascending=(method=="smallest")).ExperimentID.iloc[0]
    return j.dropna().groupby("SampleID").apply(fn)

def get_labels_new():
    E = map_experiments_to_samples()
    E = dict(zip(E.index, E))

    T = tissue()
    D = disease_state()
    columns = ["PlatformID", "ExperimentID", "TaxonID", "TissueID"] + list(D.columns)

    mdb = wrenlab.ncbi.geo.metadb.connect()
    o = pd.concat([mdb.samples, T, D], axis=1)
    o["ExperimentID"] = [E.get(id) for id in o.index.get_level_values(0)]
    o = o.loc[:,columns]
    o = o\
            .dropna(how="any", subset=o.columns[:4])\
            .dropna(how="all", subset=o.columns[4:])
    for c in ["ExperimentID", "TissueID"]:
        o[c] = [int(x) if not np.isnan(x) else None for x in o[c]]
        o[c] = o[c].astype("category")
    return o

@memoize
def get_labels():
    BTO = wrenlab.ontology.fetch("BTO")

    path = os.path.join(os.path.dirname(__file__), "labels.tsv")
    A = pd.read_csv(path, sep="\t")\
            .query("channel == 1")\
            .set_index(["sample_id"])
    A = A.query("channel == 1").copy()
    A.index = ["GSM{}".format(id) for id in A.index]
    A.index.name = "SampleID"
    A["TissueID"] = [int(x[4:]) if isinstance(x, str) else None for x in A["tissue_id"]]

    PBMC = ["peripheral blood mononuclear cell", "peripheral blood"]
    for ix in BTO.terms.index[BTO.terms["Name"].isin(PBMC)]:
        ix = int(ix[4:])
        A.ix[A["TissueID"] == ix, "TissueID"] = 89
    tissue_name_id = {name:int(id[4:]) for id,name in zip(BTO.terms.index, BTO.terms["Name"])}
    for tissue_name in EXCLUDE_TISSUES:
        tissue_id = tissue_name_id[tissue_name]
        A.ix[A["TissueID"] == tissue_id, "TissueID"] = None

    A = A.loc[:,["age","gender","TissueID"]]
    A.columns = ["Age", "Gender", "TissueID"]

    gconv = {"F":0.0,"M":1.0}
    A["Gender"] = [gconv.get(g, np.nan) for g in A["Gender"]]
    A["Gender"] = A["Gender"].astype(float)

    A["Age"] /= 12

    db = sqlite3.connect("/data/ncbi/geo/GEOmetadb.sqlite")
    # FIXME: order by biggest experiments
    gsm_gse = {k:int(v[3:]) for k,v in \
            pd.read_sql("select gsm,gse from gse_gsm;", db).to_records(index=False)}
    gsm_gpl = {k:int(v[3:]) for k,v in \
            pd.read_sql("select gsm,gpl from gsm;", db).to_records(index=False)}
    A["TaxonID"] = 9606 # FIXME
    A["ExperimentID"] = [gsm_gse.get(ix) for ix in A.index]
    A["PlatformID"] = [gsm_gpl.get(ix) for ix in A.index]
    A["Age"] = A["Age"].astype(float)

    A = A.dropna(subset=["TissueID", "ExperimentID", "PlatformID"])
    for k in ["TissueID", "ExperimentID", "PlatformID"]:
        A[k] = A[k].astype(int).astype("category")

    A.index = [int(x[3:]) for x in A.index]
    A.index.name = "SampleID"
    return A



import glob
import os
import shutil
import sqlite3

import wrenlab.util.aspera

from wrenlab.util import LOG

SRAMETADB = [
    "http://gbnci.abcc.ncifcrf.gov/backup/SRAmetadb.sqlite.gz",
    "http://watson.nci.nih.gov/~zhujack/SRAmetadb.sqlite.gz",
    "http://dl.dropbox.com/u/51653511/SRAmetadb.sqlite.gz"
]

REFERENCE = {
    "GRCh38.p7": "ftp://ftp.ncbi.nlm.nih.gov/genomes/all/GCF_000001405.33_GRCh38.p7/GCF_000001405.33_GRCh38.p7_genomic.fna.gz"
}


def reference(key):
    pass

class ReferenceGenome(object):
    def __init__(self, path):
        pass

class Aligner(object):
    def align(self, ):
        raise NotImplementedError

class Bowtie2Aligner(Aligner):
    pass


class SeqDB(object):
    def __init__(self, root):
        self._root = root
        os.makedirs(root, exist_ok=True)

        self._metadb_connect()

    def _metadb_connect(self):
        path = str(wrenlab.util.download(SRAMETADB[0], decompress=True))
        self._cx = sqlite3.connect(path)

    def __del__(self):
        self._cx.close()

    def sync(self):
        c = self._cx.cursor()
        c.execute("""
        SELECT DISTINCT(sample_accession)
            FROM sra
            WHERE 
                taxon_id=9606
                AND library_source="TRANSCRIPTOMIC"
                AND library_strategy="RNA-Seq"
                AND library_selection="RANDOM"--polyA
            ;""")

        o = os.path.join(self._root, "reads")
        os.makedirs(o, exist_ok=True)

        base = "/sra/sra-instant/reads/BySample/sra/"
        client = wrenlab.util.aspera.Client(host="ftp-trace.ncbi.nlm.nih.gov")

        for row in c:
            id = row[0]
            src = "/".join([base, id[:3], id[:6], id])
            target = os.path.join(o, id)
            try:
                client.download(src, target)
                LOG.info("Download success: SRA sample {}".format(id))
            except:
                LOG.error("Download failed: SRA sample {}".format(id))

            for file in glob.glob(os.path.join(target, "*", "*.sra")):
                shutil.move(file, target)
                shutil.rmtree(os.path.dirname(file))

    def align(self, index_name, index):
        reads_dir = os.path.join(self._root, "reads")
        for sample_id in os.listdir(reads_dir):
            sample_dir = os.path.join(reads_dir, sample_id)



import numpy as np
cimport numpy as np

from cpython cimport bool

cdef class Node:
    cdef readonly:
        int index
        str tag

    cdef bool _is_terminal(self):
        raise NotImplementedError

    @property
    def is_terminal(self):
        return self._is_terminal()

cdef class Token(Node):
    cdef readonly str word

    def __init__(self, str word, str tag=None, int index=-1):
        self.index = index
        self.word = word
        self.tag = tag

    def __str__(self):
        return "%s/%s" % (self.word, self.tag)

    def __repr__(self):
        return "<Token %s>" % str(self)

    cdef bool _is_terminal(self):
        return True

cdef class InternalNode(Node):
    cdef readonly object children

    def __init__(self, int index, str tag, 
            np.ndarray[dtype=object, ndim=1] children):
        self.index = index
        self.tag = tag
        self.children = children

    @property
    def nodes(self):
        yield self
        cdef Node child
        for child in self.children:
            if not isinstance(child, Token):
                yield from child.nodes
                pass
            else:
                yield child

    cdef bool _is_terminal(self):
        return False

    @staticmethod
    def _from_list(tree, int index=0):
        if len(tree) == 2 and isinstance(tree[1], str):
            return Token(tree[1], tree[0], index=index)
        else:
            children = []
            for i,c in enumerate(tree[1:]):
                children.append(InternalNode._from_list(c, index=index+i))
            children = np.array(children, dtype=np.object_)
            return InternalNode(index, tree[0], children)

    def __repr__(self):
        return "<InternalNode %s>" % self.tag

cdef class Tree:
    cdef InternalNode root

    def __init__(self, InternalNode root):
        self.root = root

    @property
    def tokens(self):
        cdef Node node
        for node in self.nodes:
            if isinstance(node, Token):
                yield node

    @property
    def nodes(self):
        yield from self.root.nodes

    def __iter__(self):
        return self.tokens

    def __len__(self):
        return len(self.tokens)

    @staticmethod
    def from_list(tree):
        return Tree(InternalNode._from_list(tree))



import os
import pandas as pd

import wrenlab.ncbi.gene
from wrenlab.util import memoize

@memoize
def NoaLTP():
    """
    Noa (from Zoltan Ungvari's lab) manually annotated a list of mouse genes as
    "pro" or "anti"-LTP.

    Returns
    -------
    A :class:`pandas.Series` indexed with Entrez Gene IDs and with integer values in {-1,1},
    indicating anti- or pro-LTP genes, respectively.
    """
    path = os.path.join(os.path.dirname(__file__), "NoaLTP.xlsx")
    df = pd.read_excel(path)

    genes = wrenlab.ncbi.gene.info(10090)
    symbol_id = dict(list(map(reversed, genes.loc[:,["Symbol"]].to_records())))

    df = df.loc[:,["Abbrev. ", "PRO vs ANTI LTP"]].dropna(how="any")
    df = df.ix[df.iloc[:,1].isin(["PRO", "ANTI"]),:]
    df = df.ix[df.iloc[:,0].isin(symbol_id.keys()),:]

    return pd.Series(dict([(symbol_id[k],int(v == "PRO")) for k,v in df.to_records(index=False)]))




"""
Parser for MEDLINE XML files.

Example:

.. code-block:: python
   
    from wrenlab.ncbi.medline import parse

    path = "http://www.nlm.nih.gov/databases/dtd/medsamp2013.xml.gz"
    with parse(path) as h:
        for article in path:
            print(article)
"""

import datetime
import locale
import re
import os
import multiprocessing as mp
import gzip
import pickle
import xml.etree.ElementTree as ET

from collections import namedtuple

__all__ = ["parse", "Article", "Journal"]

Article = namedtuple("Article", 
                     "id title abstract publication_date journal")
Journal = namedtuple("Journal", "id issn name")

class MedlineXMLFile(object):
    # FIXME: Date parsing will probably only work if system
    #   locale is US English

    _months = dict((i, locale.nl_langinfo(getattr(locale, "ABMON_" + str(i))))
                        for i in range(1,13))
    _non_digit_regex = re.compile(r"[^\d]+")

    def __init__(self, path):
        self._is_open = True
        self._handle = gzip.open(path, "rb")
    
    def __del__(self):
        self.close()

    def __enter__(self):
        return self

    def __exit__(self, *args, **kwargs):
        self.close()

    def close(self):
        if self._is_open:
            self._handle.close()
            self._is_open = False
    
    def _text(self, xpath):
        try:
            return self._current_element.findall(xpath)[0].text
        except IndexError:
            return None

    def _strip_non_digit(self, text):
        return self._non_digit_regex.sub('', text)

    def _parse_citation(self):
        # Parse Article information
        pmid = int(self._text(".//PMID"))
        title = self._text(".//Article/ArticleTitle")
        abstract = self._text(".//Article/Abstract/AbstractText")

        publication_date = None
        year = self._text(".//Article/Journal/JournalIssue/PubDate/Year")
        if year:
            month = self._text(".//Article/Journal/JournalIssue/PubDate/Month")
            month = self._months.get(month, "01")
            day = self._text(".//Article/Journal/JournalIssue/PubDate/Day") or "01"
            publication_date = datetime.date(int(year), int(month), int(day))

        # Parse Journal information
        journal_id = self._text(".//MedlineJournalInfo/NlmUniqueID")
        journal_id = int(self._strip_non_digit(journal_id))
        journal_issn = self._text(".//MedlineJournalInfo/ISSNLinking")
        journal_name = self._text(".//MedlineJournalInfo/MedlineTA")
        journal = Journal(journal_id, journal_issn, journal_name)

        return Article(pmid, title, abstract, publication_date, journal)

    def __iter__(self):
        for event, element in ET.iterparse(self._handle):
            if event == "end" and element.tag == "MedlineCitation":
                self._current_element = element
                yield self._parse_citation()

def parse(path_or_handle, lazy=False):
    o = MedlineXMLFile(path_or_handle)
    if lazy:
        return o
    else:
        return list(o)

def parse_all(medline_dir):
    pool = mp.Pool(mp.cpu_count() - 2)
    paths = [os.path.join(medline_dir, p) 
            for p in os.listdir(medline_dir) if p.endswith(".xml.gz")]
    seen = set()
    for articles in pool.imap(parse, paths):
        for article in articles:
            if article.id in seen:
                continue
            seen.add(article.id)
            yield article

if __name__ == "__main__":
    import argparse

    parser = argparse.ArgumentParser()
    parser.add_argument("medline_xml_file", nargs="+")
    args = parser.parse_args()
    
    for path in args.medline_xml_file:
        with parse(path) as articles:
            for article in articles:
                if not article.title:
                    continue
                text = article.title
                if article.abstract:
                    text = " ".join([text, article.abstract])
                print(article.id, text, sep="\t")



import functools

import numpy as np
import pandas as pd

from .net import download
from .memoize import memoize
from .kvstore import KVStore
from .log import LOG
from .pandas import *

###################
# Utility functions
###################

def to(t, x):
    try:
        return t(x)
    except:
        return np.nan

def as_float(x):
    return to(float, x)

def chunks(it, size=1000):
    """
    Divide an iterator into chunks of specified size. A
    chunk size of 0 means no maximum chunk size, and will
    therefore return a single list.
    
    Returns a generator of lists.
    """
    if size == 0:
        yield list(it)
    else:
        chunk = []
        for elem in it:
            chunk.append(elem)
            if len(chunk) == size:
                yield chunk
                chunk = []
        if chunk:
            yield chunk

def matrix_transformer(fn):
    """
    A decorator for a function that takes a 2D ndarray or DataFrame as
    its only positional argument (kwargs are allowed) and returns a 
    2D ndarray of the same size.

    If a :class:`pandas.DataFrame` was supplied, this decorator will
    wrap the output 2D array by adding the appropriate labels.
    """
    @functools.wraps(fn)
    def wrap(Xi, **kwargs):
        assert len(Xi.shape) == 2
        if isinstance(Xi, pd.DataFrame):
            Xim = np.array(Xi)
            Xo = fn(Xim, **kwargs)
            assert (Xi.shape == Xo.shape)
            o = pd.DataFrame(Xo, index=Xi.index, columns=Xi.columns)
            o.index.name = Xi.index.name
            o.columns.name = Xi.columns.name
            return o
        elif isinstance(Xi, np.ndarray):
            return fn(Xi, **kwargs)
        else:
            raise ValueError
    return wrap

def ranks(xs):
    """
    Return a zero-indexed array containing the ranks of each item in xs, 
    ordered from smallest to largest.

    Arguments
    ---------
    xs : array-like, 1D

    Returns
    -------
    A :class:`numpy.ndarray` containing the ranks of each item in xs, 
    ordered from smallest to largest.
    """
    o = np.empty(xs.shape[0], dtype=np.uint64)
    o[np.argsort(xs)] = np.arange(xs.shape[0], dtype=np.uint64)
    return o

class CaseInsensitiveDict(dict):
    def __setitem__(self, key, value):
        key = key.lower()
        dict.__setitem__(self, key, value)

    def __getitem__(self, key):
        key = key.lower()
        return dict.__getitem__(self, key)



"""
A read-only, disk-based, memory-mapped format for
labeled matrices (e.g., numeric data with row and column names).
"""

__all__ = ["MemoryMappedMatrix", "MMAT"]

import collections
import os
import pickle
import subprocess as sp
import sys
import warnings
import itertools
import time
import logging

from functools import partialmethod

import dask.array
import numpy as np
import pandas as pd
from scipy.stats import pearsonr

logging.basicConfig(level=logging.DEBUG)
LOG = logging.getLogger(__name__)

def format_message(msg):
    return "\n".join([line.strip() for line in msg.split("\n")
        if line.strip()])

def as_float(x):
    try:
        return float(x)
    except ValueError:
        return np.nan

def chunks(it, size=1000):
    """
    Divide an iterator into chunks of specified size. A
    chunk size of 0 means no maximum chunk size, and will
    therefore return a single list.
    
    Returns a generator of lists.
    """
    if size == 0:
        yield list(it)
    else:
        chunk = []
        for elem in it:
            chunk.append(elem)
            if len(chunk) == size:
                yield chunk
                chunk = []
        if chunk:
            yield chunk

HeaderType = np.dtype([
    ("magic", np.int64),
    ("version", np.int32),
    ("n_rows", np.int64),
    ("n_columns", np.int64),
    ("max_rows", np.int64),
    ("max_columns", np.int64),
    ("dtype_length", np.int64),
    ("loading_time", np.float32),
    ("reserved", "S65")
])

StringType = np.dtype("U100")

def slice_to_array(sl, length):
    start = sl.start or 0
    stop = sl.stop or length
    step = sl.step or 1
    return np.arange(start, stop, step)

def fix_index(ix):
    if ix is None:
        return None
    elif ix.dtype == np.object:
        ix = list(map(str, ix))
        dtype = StringType
    else:
        dtype = ix.dtype
    assert len(ix) == len(set(ix)), "duplicate elements in row or column names are not allowed."
    return np.array(ix, dtype=dtype)

def _describe(X, ix, axis):
    if axis == 1:
        return X.loc[ix,:].describe()
    elif axis == 0:
        return X.loc[:,ix].describe()
    else:
        raise ValueError

class IndexedMmap(np.memmap):
    def __init__(self, *args, **kwargs):
        super(IndexedMmap, self).__init__()
        self._ix = {}
        self._make_index()

    def _make_index(self):
        unique_elements = set(k for k,v in collections.Counter(self).items() \
                if v==1)
        self._ix = dict((v,i) for (i,v) in enumerate(self) \
                if v in unique_elements)
    
    def _get_locs(self, arg):    
        if isinstance(arg, slice):
            # A slice of index elements
            return True, slice_to_array(arg, self.shape[0])
        elif isinstance(arg, str):
            # A single string item
            return False, self._ix[arg]
        elif hasattr(arg, "__iter__"):
            # A collection of index elements
            return True, [self._ix[e] for e in arg]
        else:
            # A single item
            return False, self._ix[arg]

    def __setitem__(self, *args, **kwargs):
        super(IndexedMmap, self).__setitem__(*args, **kwargs)
        self._make_index()

class Indexer(object):
    def __init__(self, index, columns, data):
        self._index = index
        self._columns = columns
        self._data = data

    def __getitem__(self, slices):
        row_ix, col_ix = slices
        row_multi, row_ix = self._index._get_locs(row_ix)
        col_multi, col_ix = self._columns._get_locs(col_ix)
        if (not row_multi) and (not col_multi):
            return self._data[row_ix,col_ix]
        elif row_multi and col_multi:
            return View(self._index[row_ix], self._columns[col_ix],
                    self._data[row_ix,:][:,col_ix])
        elif row_multi:
            return pd.Series(self._data[row_ix,:][:,col_ix], 
                    index=self._index[row_ix],
                    name=self._columns[col_ix])
        elif col_multi:
            return pd.Series(self._data[row_ix,:][col_ix],
                    index=self._columns[col_ix],
                    name=self._index[row_ix])

class View(object):
    def __init__(self, index, columns, data):
        self.index = index
        self.columns = columns
        self.data = data

    @property
    def shape(self):
        return (self.index.shape[0], self.columns.shape[0])

    @property
    def loc(self):
        return Indexer(self.index, self.columns, self.data)

    def head(self):
        n = min(10, self.shape[0])
        return View(self.index[:n], self.columns, self.data[:n,:])

    def __repr__(self):
        return self.data.__repr__()

    def to_frame(self):
        """
        Return this View as an (in-RAM) pandas.DataFrame.
        """
        return pd.DataFrame(self.data, 
                index=self.index, 
                columns=self.columns)

    def to_dask(self):
        """
        Return this View as a :class:`dask.array.core.Array`.
        """
        return dask.array.from_array(self.data, 
                chunks=(len(self.index),1))

    def corrwith(self, series):
        assert (series.index == self.columns).all()
        series_nan = np.isnan(series)
        o = np.zeros(self.shape[0])
        o[:] = np.nan
        for i in range(self.shape[0]):
            x = self.data[i,:]
            ix = ~(series_nan | np.isnan(x))
            if ix.sum() > (0.25 * len(series)):
                o[i] = pearsonr(series.loc[ix], x[ix.nonzero()])[0]
        name = series.name or None
        return pd.Series(o, index=self.index, name=name)

    def describe(self, axis=1):
        if axis == 1:
            rows = [self.loc[ix,:].describe() for ix in self.index]
            o = pd.DataFrame(rows, index=self.index)
        elif axis == 0:
            rows = [self.loc[:,ix].describe() for ix in self.columns]
            o = pd.DataFrame(rows, index=self.columns)
        else:
            raise ValueError("'axis' must be in (0,1)")
        for c in o.columns:
            o[c] = o[c].astype(float)
        return o

    ############
    # Reductions
    ############

    # Unlike accessing View.data directly, these mask nans
    # and return pandas.Series objects 

    def reduce(self, reduce_fn, sample=None, 
            ignore_na=True, axis=0):
        assert axis in (0,1)

        if axis == 0:
            gen = (self.data[i,:] for i in range(self.shape[0]))
            ix = self.index
        else:
            gen = (self.data[:,j] for j in range(self.shape[1]))
            ix = self.columns
        o = np.empty(ix.shape[0])
        o[:] = np.nan
        for i,x in enumerate(gen):
            if ignore_na:
                mask = ~np.isnan(x)
            else:
                mask = np.ones(len(x)).astype(bool)
            if mask.sum() > 1:
                o[i] = reduce_fn(x[mask])
        return pd.Series(o, index=ix)

    def to_tsv(self, transpose=False, 
            delimiter="\t",
            file=sys.stdout):
        # FIXME: implement index.name and columns.name
        print(delimiter, end="", file=file)
        if not transpose:
            print(*self.columns, sep=delimiter, file=file)
            for i in range(self.shape[0]):
                print(self.index[i], *self.data[i,:], 
                        sep=delimiter, file=file)
        else:
            print(*self.index, sep=delimiter, file=file)
            for j in range(self.shape[1]):
                print(self.columns[j], *self.data[:,j],
                        sep=delimiter, file=file)

reductions = ["sum", "mean", "std", "var"]
for k in reductions:
    setattr(View, k, partialmethod(View.reduce, getattr(np, k)))

class MemoryMappedMatrix(View):
    """
    A simple wrapper for a 2D memory-mapped matrix with row and column IDs 
    (must be fixed-width numpy dtypes).

    The memory layout is as follows:

        byte 0-63: header, consisting of:
            byte 0-7: magic number (int64 : 0x11da549e21c7ef21)
            byte 8-12: file format version number (int32)
            byte 13-21: number of rows (int64)
            byte 22-30: number of columns (int64)
            byte 31-39: maximum number of rows (int64)
            byte 40-48: maximum number of columns (int64)
            byte 49-57: length of pickled dtype data (int64)
            byte 58-62: loading time in seconds (float32)
            byte 63-128: reserved for future expansion

        byte 129-??: pickled dtype data 
            A pickled tuple of 3 dtype objects, 
                corresponding the the dtypes of the 
                index, columns, and data. The length of 
                this byte string is specified in the header.

        byte ??-??: row index data of length equal to the 
            maximum number of rows * index dtype size
        byte ??-??: column index data of length equal to the 
            maximum number of columns * column dtype size
        byte ??-??: matrix data of size equal to the data dtype 
            * number of rows * number of columns
    """

    MAGIC_NUMBER = 0x11da549e21c7ef21
    DTYPE_OFFSET = HeaderType.itemsize

    def __init__(self, path, 
            shape=(0,0), max_shape=(1000000,1000000), 
            dtype=np.float32, 
            exclusive=False,
            index=None, columns=None,
            index_dtype=np.int64, columns_dtype=np.int64):
        """
        - exclusive : This object will be the exclusive owner 
            of the memory map. This is required if you want 
            to resize the array.
        """
        # FIXME: use portalock.lock(handle, LOCK_EX) if exclusive
        # (but requires to keep a handle around)
        self._exclusive = exclusive

        try:
            vmem = sp.check_output("ulimit -v", shell=True).strip()
            vmem = np.inf if vmem==b"unlimited" else int(vmem)
        except Exception as e:
            msg = format_message("""
            Failed to determine OS virtual memory limit
            (command 'ulimit -v' failed). If your OS is not 
            configured with a sufficiently large virtual memory 
            limit, mapping may fail. For best results,
            set this value to unlimited in your platform's
            equivalent of /etc/security/limits.conf""")
            warnings.warn(msg)
            # Proceed as though we have unlimited virtual memory
            vmem = np.inf

        self._path = path
        create = not os.path.exists(path)

        if create:
            self._map_header(create=True)
            index_data = fix_index(index)
            columns_data = fix_index(columns)
            shape = list(shape)
            if index_data is not None:
                shape[0] = len(index_data)
            if columns_data is not None:
                shape[1] = len(columns_data)

            self._header["magic"] = self.MAGIC_NUMBER
            self._header["n_rows"] = shape[0]
            self._header["n_columns"] = shape[1]
            self._header["max_rows"] = max_shape[0]
            self._header["max_columns"] = max_shape[1]

            if index_data is not None:
                index_dtype = index_data.dtype
            if columns_data is not None:
                columns_dtype = columns_data.dtype
            dtypes = (index_dtype, columns_dtype, dtype)
            dtype_str = pickle.dumps(dtypes)
            self._header["dtype_length"] = len(dtype_str)
            with open(path, "rb+") as h:
                h.seek(self.DTYPE_OFFSET)
                h.write(dtype_str)

        self.refresh()

        if create:
            if index_data is not None:
                self.index[:] = index_data
            if columns_data is not None:
                self.columns[:] = columns_data
        
    def _map_header(self, create=False):
        self._header_map = np.memmap(self._path, 
                mode="w+" if create else "r+", 
                shape=(1,), 
                dtype=HeaderType)
        self._header = self._header_map[0]

    def status(self, file=sys.stdout):
        print("loading time\t%0.3f s" 
                % self._header["loading_time"])
        print("rows", self.shape[0], sep="\t")
        print("columns", self.shape[1], sep="\t")
        print("dtype", str(self.data.dtype), sep="\t")
        print("index dtype", str(self.index.dtype), sep="\t")
        print("columns dtype", str(self.columns.dtype), sep="\t")

    def refresh(self):
        self._map_header()
        nrow = self._header["n_rows"]
        ncol = self._header["n_columns"]

        assert self._header["magic"] == self.MAGIC_NUMBER, \
                "File '%s' is not a valid MemoryMappedMatrix file."

        with open(self._path, "rb") as h:
            h.seek(self.DTYPE_OFFSET)
            index_dtype, columns_dtype, dtype = \
                pickle.loads(h.read(self._header["dtype_length"]))

        self._index_offset = \
            self.DTYPE_OFFSET + self._header["dtype_length"]
        index = IndexedMmap(self._path, 
                mode="r+", offset=self._index_offset, 
                shape=(nrow,), dtype=index_dtype)

        self._columns_offset = self._index_offset + \
                np.dtype(index_dtype).itemsize * self.max_shape[0]
        columns = IndexedMmap(self._path, 
                mode="r+", offset=self._columns_offset, 
                shape=(ncol,), dtype=columns_dtype)

        self._data_offset = self._columns_offset + \
            np.dtype(columns_dtype).itemsize * self.max_shape[1]
        data = np.memmap(self._path, 
                mode="r+", offset=self._data_offset, 
                shape=(nrow, ncol), dtype=dtype)

        self.dtype = dtype
        super(MemoryMappedMatrix, self)\
                .__init__(index, columns, data)

    @property
    def max_shape(self):
        return (self._header["max_rows"], 
                self._header["max_columns"])

    def resize(self, shape):
        assert self._exclusive

        nrow, ncol = shape
        assert nrow <= self.max_shape[0]
        assert ncol <= self.max_shape[1]

        self._header["n_rows"] = nrow
        self._header["n_columns"] = ncol

        self.refresh()

        self.index._make_index()
        self.columns._make_index()

    def flush(self):
        self._header_map.flush()
        self.data.flush()
        self.index.flush()
        self.columns.flush()

    def __del__(self):
        self.flush()


    @staticmethod
    def from_file(handle, path, delimiter="\t"):
        LOG.info("Beginning data import to: {}".format(path))

        chunk_size = 100
        growth_factor = 2
        start_time = time.time()

        def maybe_int(items):
            try:
                items = list(map(int, items))
            except ValueError:
                pass
            return pd.Index(items)

        with handle:
            columns = maybe_int(next(handle)\
                    .rstrip("\n").split(delimiter)[1:])
            nc = len(columns)
            #cnks = chunks(handle, chunk_size)
            #chunk = next(cnks)
            line = next(handle)
            try:
                int(line.split(delimiter)[0])
                index_dtype = np.int64
            except ValueError:
                index_dtype = StringType
            lines = itertools.chain([line], handle)

            LOG.info("Column count: {}".format(len(columns)))
            X = MMAT(path, 
                    shape=(100, len(columns)),
                    dtype=np.float32,
                    max_shape=(1000000,nc),
                    exclusive=True,
                    index_dtype=index_dtype,
                    columns=columns)

            index = []

            nr = 0
            for nr,line in enumerate(lines):
                if nr + 1 > X.shape[0]:
                    X.resize((int(X.shape[0]*growth_factor), 
                        nc))
                    LOG.info("Processed row: {}".format(nr))
                isep = line.find(delimiter)
                ix = line[:isep]
                if ix in index:
                    raise ValueError("Duplicate index name: {}".format(ix))
                index.append(ix)
                #X.data[nr,:] = np.fromstring(line[isep+1:], sep=delimiter)
                X.data[nr,:] = list(map(as_float, line.rstrip("\n").split(delimiter)[1:]))

            X.resize((nr+1,nc))
            X.index[:] = fix_index(maybe_int(index))
            X._header["loading_time"] = time.time() - start_time
            LOG.info("Import complete")
            return X

MMAT = MemoryMappedMatrix

########################
# Command-line interface
########################

import click

@click.group()
def cli():
    pass

@cli.command()
@click.argument("matrix_path", required=True)
@click.option("--delimiter", "-d", default="\t")
def load(matrix_path, delimiter="\t"):
    """
    Create a MMAT from TSV matrix
    """
    import sys
    X = MMAT.from_file(sys.stdin, matrix_path, delimiter=delimiter)
    start = time.time()
    e = lambda *args: print(*args, file=sys.stderr)
    e("\nMatrix loaded:")
    X.status()

@cli.command()
@click.argument("matrix_path", required=True)
@click.option("--delimiter", "-d", default="\t")
@click.option("--transpose/--no-transpose", "-t", default=False)
def dump(matrix_path, delimiter="\t", transpose=False):    
    """
    Export a matrix's data to TSV
    """
    X = MMAT(matrix_path)
    X.to_tsv(delimiter=delimiter, transpose=transpose)

@cli.command()
@click.argument("matrix_path", required=True)
def status(matrix_path, delimiter="\t"):    
    """
    Print MMAT shape, dtype, etc
    """
    X = MMAT(matrix_path)
    X.status()

@cli.command()
@click.argument("matrix_path", required=True)
@click.argument("reduction", required=True)
@click.option("--axis", "-a", type=int, default=0)
def reduce(matrix_path, reduction, axis=0):
    """
    Reduce the matrix along an axis
    """
    X = MMAT(matrix_path)
    fn = getattr(np, reduction)
    X.reduce(fn, axis=axis)\
            .to_csv(sys.stdout, sep="\t", na_rep="nan")

@cli.command()
@click.argument("matrix_path", required=True)
@click.option("--ignore-missing", "-i",
        help="Don't throw an error if a label is missing, just omit it",
        is_flag=True)
@click.option("--rows", "-r", 
        help="Input lines are row labels (default)",
        is_flag=True)
@click.option("--columns", "-c", 
        help="Input lines are column labels",
        is_flag=True)
def select(matrix_path, rows, columns, ignore_missing):
    """
    Select rows or columns by key
    """
    if not rows or columns:
        rows = True
    assert (rows != columns), "-r and -c are mutually exclusive"
    X = MMAT(matrix_path)
    selection = [line.rstrip("\n") for line in sys.stdin]
    if ignore_missing:
        ix = set(X.index if rows else X.columns)
        selection = [x for x in selection if x in ix]
    if rows:
        X.loc[selection,:].to_tsv()
    else:
        X.loc[:,selection].to_tsv()

@cli.command()
@click.argument("matrix_path", required=True)
@click.option("--rows", "-r", is_flag=True)
@click.option("--columns", "-c", is_flag=True)
def keys(matrix_path, rows, columns):
    """
    Output row or column names
    """
    if not (rows or columns):
        rows = True
    assert (rows != columns), "-r and -c are mutually exclusive"
    X = MMAT(matrix_path)
    if rows:
        keys = X.index
    else:
        keys = X.columns
    for k in keys:
        print(k)
 
if __name__ == "__main__":
    cli()

def test():
    X = MemoryMappedMatrix("/home/gilesc/test.mmat", shape=(10,10))
    X.columns[:] = np.arange(10) 
    X.index[:] = np.arange(10) 
    print(dir(X.data))

    print(X.loc[3:6,3])
    X.resize((20,20))
    print(X)



"""
Matrix normalization methods (primarily for expression data).
"""

import numpy as np
import pandas as pd

from wrenlab.util import matrix_transformer

def vsn(X):
    """
    Variance-stabilizing normalization.

    Arguments
    ---------
    X : :class:`pandas.DataFrame`
        An expression matrix, with probes as rows and samples as columns.

    Returns
    -------
    A :class:`pandas.DataFrame` containing the VSN normalized 
    expression values.
    """
    import wrenlab.R
    return wrenlab.R.vsn(X)

def vsn2(X):
    import wrenlab.R
    params = wrenlab.R.fit_rocke_durbin(X, theta=0.5)
    print(params)
    alpha = params["alpha"]
    v_eta = params["sd_eta"] ** 2
    v_epsilon = params["sd_epsilon"] ** 2
    S2_eta = np.exp(v_eta) * (np.exp(v_eta) - 1)
    c = v_epsilon / S2_eta

    Xm = np.array(X)
    return pd.DataFrame(np.log(Xm - alpha + np.sqrt((Xm - alpha)**2 + c)),
            index=X.index, columns=X.columns)

@matrix_transformer
def quantile(X, mu=None):
    """
    Quantile normalize a matrix.

    Parameters
    ----------
    X : a 2D :class:`numpy.ndarray`
        The matrix to be normalized, with samples as columns
        and probes/genes as rows.
    mu : a 1D :class:`numpy.ndarray`, optional
        Vector of gene means.

    Returns
    -------
    :class:`numpy.ndarray`
        The normalized matrix.
    """
    # transposed, so samples are rows
    assert not np.isnan(X).all(axis=1).any() # rows
    if mu is not None:
        mu = np.array(mu)
        assert len(mu) == X.shape[1] 
        assert not np.isnan(mu).any()
    else:
        assert not np.isnan(X).all(axis=0).any() # columns

    Xm = np.ma.masked_invalid(X.T)
    Xn = np.empty(Xm.shape)
    Xn[:] = np.nan

    if mu is None:
        mu = Xm.mean(axis=0)
    mu.sort()

    for i in range(Xm.shape[0]):
        # sort and argsort sorts small to large with NaN at the end
        ix = np.argsort(Xm[i,:])
        nok = (~Xm[i,:].mask).sum()
        ix = ix[:nok]
        rix = (np.arange(nok) * len(mu) / nok).round().astype(int)
        Xn[i,ix] = mu[rix]

    Xn = Xn.T
    assert (np.isnan(X) == np.isnan(Xn)).all()
    return Xn



import pandas as pd

from wrenlab.util import memoize
import wrenlab.ontology

@memoize
def concepts():
    o = wrenlab.ontology.fetch("GO")
    ts = o.terms
    ix = ts["Name"].str.match("^(?:positive|negative) regulation of",
            as_indexer=True)
    coef = pd.Series(0, index=ts.index)
    coef[ts["Name"].str.match("^positive regulation of")] = 1
    coef[ts["Name"].str.match("^negative regulation of")] = -1
    ts["Concept"] = ts["Name"].str.replace(\
        "^(?:(?:positive|negative) )?regulation of ", "")
    ts["Directionality"] = coef
    o = []
    for concept, df in ts.groupby("Concept"):
        try:
            pos = df.ix[df["Directionality"] == 1,:].index[0]
            neg = df.ix[df["Directionality"] == -1,:].index[0]
        except IndexError:
            continue
        o.append((concept, pos, 1))
        o.append((concept, neg, -1))
    return pd.DataFrame.from_records(o, columns=["Concept", "Term ID", "Direction"])



import functools

import sklearn.cross_decomposition
from cython.parallel import prange
import pandas as pd
import numpy as np
cimport numpy as np
import scipy.stats

import wrenlab.util

def standardize(np.ndarray[np.float64_t, ndim=1] x):
    ix = ~np.isnan(x)
    mu, std = x[ix].mean(), x[ix].std()
    return (x - mu) / std

cdef _pearson(np.ndarray[np.float64_t, ndim=2] X, np.ndarray[np.float64_t, ndim=1] y):
    assert X.shape[1] == y.shape[0]
    ix = ~np.isnan(y)
    assert ix.sum() >= 3
    ns = X.shape[0]
    no = ix.sum()
    X = X[:,ix]

    Xs = np.zeros_like(X)
    cdef int i
    for i in range(ns):
        Xs[i,:] = standardize(X[i,:])
    Xs[np.isnan(Xs)] = 0
    ys = standardize(y[ix])

    cdef np.ndarray[np.int64_t, ndim=1] N = (~np.isnan(X)).sum(axis=1)
    cdef np.ndarray[np.float64_t, ndim=1] r = (Xs @ ys) / N
    cdef np.ndarray[np.float64_t, ndim=1] t = r * ((N - 2) ** 0.5) / (1 - r**2)
    r[N < 3] = np.nan
    t[N < 3] = np.nan

    cdef np.ndarray[np.float64_t, ndim=1] p = scipy.stats.t.cdf(t, df=N - 2)
    p[np.isclose(r, 1) | np.isclose(r, -1)] = 0
    p = np.minimum(p, 1-p) * 2
    assert (((p <= 1) & (p >= 0)) | np.isnan(p)).all()
    cdef np.ndarray[np.float64_t, ndim=1] SLPV = -1 * np.sign(t) * np.log10(p)
    return pd.DataFrame.from_records(map(tuple, zip(N,r,p,SLPV)), 
            columns=["N","r","p","SLPV"])

cdef _spearman(np.ndarray[np.float64_t, ndim=2] X, np.ndarray[np.float64_t, ndim=1] y):
    assert X.shape[1] == y.shape[0]
    ix = ~np.isnan(y)
    assert ix.sum() >= 3
    ns = X.shape[0]
    no = ix.sum()
    X = X[:,ix]

    cdef np.ndarray[np.float64_t, ndim=1] rho = np.zeros(ns)
    rho[:] = np.nan
    cdef np.ndarray[np.int64_t, ndim=1] N = (~np.isnan(X)).sum(axis=1)

    cdef int i
    for i in range(ns):
        x = X[i,:]
        n = N[i]
        if n < 3:
            continue
        xr = wrenlab.util.ranks(x[ix]) + 1
        yr = wrenlab.util.ranks(y[ix]) + 1
        rho[i] = 1 - (6 * ((xr - yr) ** 2).sum()) / (n * (n ** 2 - 1))

    cdef np.ndarray[np.float64_t, ndim=1] F = np.arctanh(rho)
    cdef np.ndarray[np.float64_t, ndim=1] z = F * np.sqrt((N-3)/1.06)
    cdef np.ndarray[np.float64_t, ndim=1] p = scipy.stats.norm.cdf(z)
    p[np.isclose(rho, 1) | np.isclose(rho, -1)] = 0
    p = np.minimum(p, 1-p) * 2
    assert (((p <= 1) & (p >= 0)) | np.isnan(p)).all()
    cdef np.ndarray[np.float64_t, ndim=1] SLPV = -1 * np.sign(rho) * np.log10(p)
    return pd.DataFrame.from_records(map(tuple, zip(N,rho,p,SLPV)), 
            columns=["N","r","p","SLPV"])

import numpy as np
from scipy import stats, linalg

def partial_correlation(
        np.ndarray[np.float64_t, ndim=1] x,
        np.ndarray[np.float64_t, ndim=1] y,
        np.ndarray[np.float64_t, ndim=2] C):
    """
    Returns the sample linear partial correlation coefficients between x and y, 
    controlling for the variables in C.

    Parameters
    ----------
    x : np.ndarray, shape (n,)
    y : np.ndarray, shape (n,)
    C : np.ndarray, shape (n,p)

    Returns
    -------
    A tuple.

    r : float
        The partial correlation coefficient between x and y.
    p : float
        The p-value for the correlation.
    """
    n = x.shape[0]
    p = C.shape[1]

    beta_x = np.linalg.lstsq(C, x)[0]
    beta_y = np.linalg.lstsq(C, y)[0]
    res_x = x - (C @ beta_x)
    res_y = y - (C @ beta_y)
    r, p = scipy.stats.pearsonr(res_x, res_y)
    return r, p
    
def spearman(X, y):
    y = np.array(y)
    if isinstance(X, np.ndarray):
        return _spearman(X, y)
    elif isinstance(X, pd.DataFrame):
        Xm = np.array(X)
        o = _spearman(Xm, y)
        o.index = X.index
        o.index.name = X.index.name
        return o
    else:
        raise TypeError
    
def pearson(X, y):
    y = np.array(y)
    if isinstance(X, np.ndarray):
        return _pearson(X, y)
    elif isinstance(X, pd.DataFrame):
        Xm = np.array(X)
        o = _pearson(Xm, y)
        o.index = X.index
        o.index.name = X.index.name
        return o
    else:
        raise TypeError

def CCA(X, Y, components=5):
    """
    Run canonical correlation analysis (CCA) between X and Y,
    returning the weight matrices.
    """
    assert isinstance(X, pd.DataFrame)
    assert isinstance(Y, pd.DataFrame)
    X, Y = X.dropna(how="any").align(Y.dropna(how="any"), axis=0, join="inner")
    X = X.apply(lambda x: standardize(np.array(x)), axis=0)
    Y = Y.apply(lambda x: standardize(np.array(x)), axis=0)

    model = sklearn.cross_decomposition.CCA(n_components=components)
    model.fit(X,Y)

    def labels(n):
        return ["C{}".format(i) for i in range(1, n+1)]

    WX = pd.DataFrame(model.x_weights_, index=X.columns, columns=labels(components))
    WY = pd.DataFrame(model.y_weights_, index=Y.columns, columns=labels(components))
    return (WX, WY)




import functools
import zipfile
import sqlite3
import contextlib

import pandas as pd

import wrenlab.ncbi.gene
import wrenlab.util

from wrenlab.data.drugbank import DrugBank

def AILUN(platform_id):
    url = "ftp://phoenix.omrf.hsc.net.ou.edu/AILUN.db"
    path = str(wrenlab.util.download(url))
    with contextlib.closing(sqlite3.connect(path)) as db:
        c = db.cursor()
        c.execute("SELECT probe_id, gene_id FROM probe_gene WHERE platform_id=?", 
                (platform_id,))
        return dict(c)

def connectivity_map_design():
    url = "https://www.broadinstitute.org/cmap/cmap_instances_02.xls"
    path = str(wrenlab.util.download(url))
    D = pd.read_excel(path)\
            .loc[:,["instance_id","batch_id","cmap_name",
                "concentration (M)", "duration (h)", "cell2", "vehicle"]]
    D.columns = ["SampleID", "BatchID", "Drug", "Concentration", 
            "Duration", "CellType", "Vehicle"]
    D = D.set_index(["SampleID"])
    D = D.ix[[isinstance(ix, int) for ix in D.index],:]
    D.index = D.index.astype(int)
    for c in D.columns:
        if c not in ("Concentration", "Duration"):
            D[c] = D[c].astype("category")
    return D

@functools.lru_cache()
def connectivity_map():
    # the data values are ranks 1 ... N
    # with rank 1 being highest treatment / control ratio and rank N being lowest
    url = "ftp://ftp.broad.mit.edu/pub/cmap/rankMatrix.txt.zip"
    path = str(wrenlab.util.download(url))
    with zipfile.ZipFile(path) as zf:
        with zf.open("rankMatrix.txt", "r") as h:
            X = pd.read_csv(h, sep="\t")\
                    .set_index(["probe_id"])\
                    .iloc[:,:-1]
            X.columns = list(map(int, X.columns))
            probe_gene = {k:v for k,v in AILUN(570).items() if k in X.index}
            X = X.loc[list(probe_gene.keys()),:]
            rows = []
            for gene_id, df in X.groupby(probe_gene):
                v = df.mean()
                v.name = gene_id
                rows.append(v)
            Xc = pd.concat(rows, axis=1)
            D = connectivity_map_design()
            D, Xc = D.align(Xc, axis=0, join="inner")
            Xc = Xc.T
            Xc.index.name = "Entrez Gene ID"
            Xc.columns.name = "SampleID"
            Xc = Xc.apply(lambda x: wrenlab.util.ranks(-x))
            #return D, Xc
            Xc = Xc.groupby(D["Drug"].astype(str), axis=1)\
                    .mean()\
                    .apply(lambda x: wrenlab.util.ranks(-x))
            return Xc

if __name__ == "__main__":
    """
    X = connectivity_map()
    mu = X.mean().sort_index(ascending=False)
    print(mu.describe())
    print(mu.head().index)
    print(mu.tail().index)
    #print(X.head().T.head().T)
    """

    #print(AILUN(570))

    #D = connectivity_map_design()
    #print(D.tail())

    #pprint.pprint(db._items)



import collections
import zipfile

import pandas as pd
import lxml.etree

import wrenlab.util
from wrenlab.util import memoize
import wrenlab.ncbi.gene

Drug = collections.namedtuple("Drug", "drugbank_id,name,synonyms,indication,targets")
Target = collections.namedtuple("Target", "gene_id,actions")

@memoize
def parse_drugbank():
    NS = "{http://www.drugbank.ca}"
    genes = wrenlab.ncbi.gene.info(9606)
    symbol_id = dict(zip(genes["Symbol"], list(map(int, genes.index))))

    def tag(key):
        return "{}{}".format(NS, key)

    def find(n, *keys):
        for k in keys:
            if n is None:
                return
            n = n.find(tag(k))
        return n

    def find_text(n, *keys):
        n = find(n, *keys)
        if n is not None:
            return n.text

    def findall(n, *keys):
        # find on all but the last key, findall on last
        keys = list(keys)
        n = find(n, *keys[:-1])
        if n is not None:
            o = n.findall(tag(keys[-1]))
            if o is not None:
                return o
        return []

    def handle_drug(e):
        drugbank_id = set()
        for n in e.findall(tag("drugbank-id")):
            drugbank_id.add(n.text)

        name = e.find(tag("name")).text

        synonyms = set()
        for n in findall(e, "synonym", "synonym"):
            synonyms.add(n.text)

        indication = e.find(tag("indication")).text

        targets = []
        for n in findall(e, "targets", "target"):
            actions = set([a.text.lower() for a in findall(n, "actions", "action")])
            symbol = find_text(n, "polypeptide", "gene-name")
            gene_id = symbol_id.get(symbol)
            if gene_id is not None:
                target = Target(gene_id, actions)
                targets.append(target)

        return Drug(drugbank_id, name, synonyms, indication, targets)

    url = "http://www.drugbank.ca/system/downloads/current/drugbank.xml.zip"
    path = wrenlab.util.download(url)
    o = []
    with zipfile.ZipFile(path) as zf:
        with zf.open("drugbank.xml") as h:
            ctx = lxml.etree.iterparse(h, events=("end",))
            for event, e in ctx:
                if (event == "end") and (e.tag == tag("drug")):
                    drug = handle_drug(e)
                    o.append(drug)
    return o


class DrugBank(object):
    def __init__(self):
        self._items = parse_drugbank()
        self._synonyms = wrenlab.util.CaseInsensitiveDict()
        for drug in self._items:
            self._synonyms[drug.name] = drug
            for s in drug.synonyms:
                self._synonyms[s] = drug

    def by_name(self, name):
        return self._synonyms.get(name)

    @property
    def targets(self):
        SIGN_MAP = {
            "agonist": 1,
            "stimulator": 1,
            "activator": 1,
            "antagonist": -1,
            "inhibitor": -1,
        }
        rows = []
        for drug in self._items:
            for target in drug.targets:
                signs = [SIGN_MAP.get(t) 
                        for a in target.actions
                        for t in a.split()
                        if t in SIGN_MAP]
                sign = signs[0] if len(signs) == 1 else 0
                rows.append((drug.name, target.gene_id, sign))
        return pd.DataFrame.from_records(rows,
                columns=["Drug", "Target", "Direction"])


if __name__ == "__main__":
    import pprint
    db = DrugBank()
    print(db.targets.head())



from collections import namedtuple
import os.path
import urllib.request
import shelve

import yaml
import pandas as pd
import numpy as np

from wrenlab.util.remote_table import remote_table
from wrenlab.util import memoize, download, KVStore

def gene_taxon():
    if not KVStore.exists("gene_taxon"):
        LOG.debug("Initializing KVStore for gene_taxon ...")
        tbl = remote_table("NCBI.gene.info")
        store = KVStore("gene_taxon", writeback=True)
        for taxon_id, gene_id in tbl.index.values:
            store[int(gene_id)] = int(taxon_id)
        store.sync()
        store.close()
    return KVStore("gene_taxon")

@memoize
def info(taxon_id):
    assert isinstance(taxon_id, int)
    genes = remote_table("NCBI.gene.info")
    return genes.loc[taxon_id,:]

@memoize
def _synonyms():
    tbl = remote_table("NCBI.gene.synonym")
    o = []
    for (taxon_id, gene_id, synonyms, designations) in tbl.to_records(index=False):
        text = "|".join([field for field in [synonyms, designations] if
            (field is not None) and (field == field)])
        for synonym in text.split("|"):
            synonym = synonym.strip()
            if synonym:
                o.append((taxon_id, gene_id, synonym))
    o = pd.DataFrame.from_records(o, columns=["Taxon ID", "Gene ID", "Synonym"])
    return o.drop_duplicates()

@memoize
def synonyms(taxon_id):
    tbl = _synonyms()
    return tbl.ix[tbl["Taxon ID"] == taxon_id,:].drop("Taxon ID", axis=1)

@memoize
def accession(taxon_id):
    assert isinstance(taxon_id, int)
    return remote_table("NCBI.gene.accession")\
            .loc[taxon_id,:]\
            .dropna(how="all")

@memoize
def ensembl(taxon_id):
    assert isinstance(taxon_id, int)
    o = remote_table("NCBI.gene.ensembl")
    return o.ix[o["Taxon ID"] == taxon_id,:]\
            .drop("Taxon ID", axis=1)

Gene = namedtuple("Gene", "id,symbol,name,chromosome")

def by_id(gene_id):
    m = gene_taxon()
    taxon_id = m[gene_id]
    df = info(taxon_id)
    symbol = df.loc[gene_id, "Symbol"]
    name = df.loc[gene_id, "Name"]
    chromosome = df.loc[gene_id, "Chromosome"]
    return Gene(gene_id, symbol, name, chromosome)



"""
Methods for imputing missing values in matrices.
"""

import math

import pandas as pd
import numpy as np

from wrenlab.util import matrix_transformer, LOG

def qPCR(X, design, controls, max_iterations=100):
    import wrenlab.R
    return wrenlab.R.impute_qPCR(X, design, controls, 
            max_iterations=max_iterations)

def kNN(X, k=None):
    """
    Impute missing values using mean of KNN.

    Arguments
    ---------
    X: :class:`pandas.DataFrame`
        The expression matrix, with probes as rows and samples as columns.

    Returns
    -------
    A :class:`pandas.DataFrame` with missing values imputed, and rows containing
    unimputable values dropped. The resulting :class:`pandas.DataFrame` will
    have no missing values and may have fewer rows.
    """
    assert isinstance(X, pd.DataFrame)
    if k is None:
        k = math.ceil(X.shape[1] * 0.1)
    assert isinstance(k, int)

    drop_threshold = int(0.25 * X.shape[1])
    X = X.dropna(axis=0, thresh=drop_threshold)
    nn = X.corr().apply(lambda r: r.sort_values(ascending=False).index).iloc[1:,:].T
    Xn = X.copy()
    for i,j in zip(*np.array(X.isnull()).nonzero()):
        mu = X.iloc[i,:].loc[nn.iloc[j,:]].dropna().iloc[:k].mean()
        Xn.iloc[i,j] = mu
    return Xn.dropna(axis=0)

@matrix_transformer
def mean(X):
    """
    Impute missing values using the mean value of the gene.

    Arguments
    ---------
    X: :class:`numpy.ndarray`
        The expression matrix, with probes as rows and samples as columns.

    Returns
    -------
    A :class:`pandas.DataFrame` with missing values imputed, and rows containing
    unimputable values dropped. The resulting :class:`pandas.DataFrame` will
    have no missing values and may have fewer rows.
    """
    Xo = X.copy()
    X = np.ma.masked_invalid(X)
    mu = X.mean(axis=1)
    for i in range(X.shape[0]):
        Xo[i,X.mask[i,:]] = mu[i]
    return Xo

@matrix_transformer
def linear_model(X, max_iterations=5, convergence=1e-8, use_subset=True):
    if use_subset:
        max_predictors = min(500, X.shape[0] - 1)
    else:
        max_predictors = None

    nt, ns = X.shape
    missing = np.isnan(X.T)
    assert not missing.all(axis=0).any()

    converged = np.zeros(nt, dtype=bool)
    converged[missing.sum(axis=0) == 0] = True
    Xi = mean(X).T

    LOG.info("Starting linear model based imputation on matrix of shape: {}".format(X.shape))
    delta_prev = 2 ** 64

    for iteration in range(max_iterations):
        Xi_next = Xi.copy()
        delta = np.zeros(nt)
        for j in (~converged).nonzero()[0]:
            ix = missing[:,j]
            train = np.delete(Xi, np.s_[j], 1)
            if max_predictors is not None:
                predictors = np.random.choice(np.arange(train.shape[1]),
                        size=max_predictors,
                        replace=False)
                train = train[:,predictors]
            beta, _, _, sv = np.linalg.lstsq(train[~ix,:], Xi[~ix,j])
            y_hat = train @ beta
            Xi_next[ix,j] = y_hat[ix]
            delta[j] = ((y_hat[ix] - Xi[ix,j]) ** 2).mean()
        converged[delta < convergence] = True
        if converged.all():
            LOG.info("Convergence after {} iterations".format(iteration + 1))
            return Xi_next.T
        else:
            LOG.debug("Iteration {}: mean delta = {}, not converged = {}"\
                    .format(
                        iteration + 1, 
                        delta[~converged].mean(), 
                        (~converged).sum()))
            Xi = Xi_next
            if (delta_prev < delta.mean()) and (max_predictors is not None):
                # 4000 predictors is where it starts to get really slow
                max_predictors = min(int(max_predictors * 1.5), 4000)
                LOG.debug("max_predictors increased to: {}".format(max_predictors))
            delta_prev = delta.mean()
    return Xi.T



import functools
import tarfile
from collections import namedtuple

import pandas as pd

from wrenlab.util import memoize, download

Taxon = namedtuple("Taxon", "id,name")

@memoize
def names():
    scientific = {}
    common = {}
    url = "ftp://ftp.ncbi.nih.gov/pub/taxonomy/taxdump.tar.gz"
    path = str(download(url))
    tar = tarfile.open(path, "r:gz")
    handle = tar.extractfile("names.dmp")
    for i,line in enumerate(handle):
        fields = [c.strip("\t") for c in
                line.decode("utf-8")\
                        .rstrip("\n").split("|")][:-1]
        taxon_id = int(fields[0])
        name = fields[1]
        unique_name = fields[2]
        name_class = fields[3]
        if name_class == "scientific name":
            scientific[taxon_id] = name
        elif "common name" in name_class:
            common[taxon_id] = name

    records = [(taxon_id, scientific[taxon_id], common.get(taxon_id)) 
            for taxon_id in scientific]
    return pd.DataFrame(records, columns=["Taxon ID", "Scientific Name", "Common Name"])\
            .set_index(["Taxon ID"])

def by_id(taxon_id):
    assert isinstance(taxon_id, int)

    df = names()
    name = df.loc[taxon_id, "Scientific Name"]
    return Taxon(taxon_id, name)



"""
Python interface to R functions.
"""

import collections
from contextlib import redirect_stdout, redirect_stderr
import os
import warnings

import pandas as pd
import numpy as np
import scipy.stats

from rpy2.robjects import r, pandas2ri
from rpy2.robjects.packages import importr

from wrenlab.util import LOG

pandas2ri.activate()

r["Sys.setenv"](R_SHARE_DIR="/usr/share/R")

def dict_to_list(d):
    ks, vs = map(list, zip(*d.items()))
    o = r["as.list"](vs)
    o.names = ks
    return o

def vsn(X):
    """
    Variance-stabilizing normalization.

    Arguments
    ---------
    X : :class:`pandas.DataFrame`
        An expression matrix, with probes as rows and samples as columns.
    """
    LOG.info("Calling R VSN function on matrix of shape: {}".format(X.shape))
    pkg = importr("vsn")
    Xm = np.array(X)
    with open(os.devnull, "w") as h:
        with redirect_stdout(h):
            o = np.array(r["as.matrix"](pkg.vsn2(Xm)))
    o = pd.DataFrame(o, index=X.index, columns=X.columns)
    o.index.name = X.index.name
    o.columns.name = X.columns.name
    return o

def impute_qPCR(X, design, controls=None, max_iterations=100):
    if controls is None:
        controls = list(X.index[~X.isnull().any(axis=1)])

    pkg = importr("wrenlab")
    c = r["unlist"](controls)
    with open(os.devnull, "w") as h:
        with redirect_stdout(h):
            with warnings.catch_warnings():
                warnings.simplefilter("ignore")
                o = pkg.impute_qPCR(X, design, c, max_iterations=max_iterations)
    o = pandas2ri.ri2py(o)
    return pd.DataFrame(o, index=X.index, columns=X.columns)

def wilcoxGST(index, statistics, alternative="mixed", type="auto"):
    assert alternative in ("mixed", "either", "up", "down")
    assert type in ("t", "f", "auto")
    # unused: ranks_only (default T), nsim
    pkg = importr("limma")
    return float(pkg.geneSetTest(index, statistics, alternative=alternative, type=type, ranks_only=True)[0])

def fit_rocke_durbin(X, theta=0.05):
    pkg = importr("GESTr")
    Xm = np.array(X)
    with open(os.devnull, "w") as h:
        with redirect_stdout(h):
            o = pkg.fitRockeDurbin(Xm, theta=theta)
    return dict(zip(o.names, [x[0] for x in o]))

def mi(X):
    """
    Multiple imputation using the "mi" package.

    Arguments
    ---------
    X : :class:`pandas.DataFrame`
        A matrix with variables as rows and samples as columns.
    """
    X = X.dropna(thresh=10)
    pkg = importr("mi")
    X_r = pkg.missing_data_frame(X.T)
    X_r = pkg.change(X_r, y=r["as.character"](["X{}".format(i) for i in X.index]), 
            what="family", to="gaussian")
    imputations = pkg.mi(X_r, **{"n.iter":30, "n.chains":4, "max.minutes": 20})
    r.show(imputations)

def compose(*fns):
    def fn(x):
        for fn in reversed(fns):
            x = r[fn](x)
        return x
    return fn

def rstrip_digit(x):
    for i in reversed(range(len(x))):
        if not x[i].isdigit():
            return x[:(i+1)]
    return ""

def fix_index(index):
    o = []
    for ix in index:
        if ix == "(Intercept)":
            o.append("Intercept")
            continue
        else:
            ixs = []
            for six in ix.split(":"):
                base = rstrip_digit(six)
                if len(base) < len(six):
                    digit = six[len(base):]
                    six = "{}[{}]".format(base, digit)
                ixs.append(six)
            o.append(":".join(ixs))
    return o

class LMERModel(object):
    def __init__(self, fit):
        self._fit = fit

    @property
    def fixed_effects(self):
        o = pandas2ri.ri2py(compose("as.data.frame", "coef", "summary")(self._fit))
        o.index = fix_index(o.index)
        o.columns = ["coef", "SE", "t"]
        o["p"] = 2 * (1 - scipy.stats.norm.cdf(o["t"].abs()))
        return o

    @property
    def random_effects(self):
        print(r["summary"](self._fit))
        re = r["ranef"](self._fit)
        o = collections.OrderedDict()
        for key, df in zip(re.names, re):
            o[key] = pandas2ri.ri2py(df)
            o[key].columns = fix_index(o[key].columns)
        return o

def lmer(formula, y, design):
    assert y.shape[0] == design.shape[0]
    assert "y" not in design.columns

    formula_r = r["as.formula"]("y ~ {}".format(formula))
    D = design.copy()
    D["y"] = y

    pkg = importr("lme4")
    fit = pkg.lmer(formula_r, D)
    return LMERModel(fit)

class StanModel(object):
    def __init__(self, path):
        pkg = importr("rstan")
        self._model = pkg.stan_model(path)

    def fit(self, data, iter=1000, chains=4):
        pkg = importr("rstan")
        data = dict_to_list(data)
        fit = self._model
        fit = pkg.sampling(self._model, data=data, iter=iter, chains=chains)
        return StanFit(fit)

    def save(self, path):
        pass

    @staticmethod
    def load(path):
        pass

class StanFit(object):
    def __init__(self, fit):
        self._fit = fit

    def serve(self, host="127.0.0.1", port=8080, launch_browser=False):
        pkg = importr("shinystan")
        pkg.launch_shinystan(object=self._fit, 
                port=port, 
                **{"launch.browser": launch_browser})

def stan(model_path, data, iter=1000, chains=4):
    pkg = importr("rstan")
    data = dict_to_list(data)
    #handle = r["file"](model_path)
    fit = pkg.stan(file=model_path, data=data, iter=iter, chains=chains)
    return StanFit(fit)



"""
Function memoization using joblib.
"""

import os
import os.path

from joblib import Memory

CACHE_DIRECTORY = os.path.expanduser("~/.cache/wrenlab/cache/")
os.makedirs(CACHE_DIRECTORY, exist_ok=True)
memoize = Memory(cachedir=CACHE_DIRECTORY, compress=True).cache



"""
Machine learning utilities.
"""

import seaborn as sns
import matplotlib.pyplot as plt
import numpy as np
import scipy.stats
import sklearn.cross_validation
import sklearn.metrics
import pandas as pd

from wrenlab.util import align_series

class Problem(object):
    def __init__(self, y, X, type=None):
        assert isinstance(self.y, pd.Series)

        self.y = y
        self.X = X
        assert y.shape[0] == X.shape[0]
        if len(set(y)) == 2:
            self.type = "binary"
            self.predict_fn = lambda model, X: model.predict_log_proba(X)
        else:
            self.type = "continuous"
            self.predict_fn = lambda model, X: model.predict(X)

    def cross_validate(self, model, k=10):
        X = np.array(self.X)
        y = np.array(self.y)
        kf = sklearn.cross_validation.KFold(y.shape[0], k)
        y_hat = np.empty(y.shape)
        for tr_ix, te_ix in kf:
            X_tr, y_tr = X[tr_ix,:], y[tr_ix]
            model.fit(X_tr, y_tr)
            y_hat[te_ix] = self.predict_fn(model, X[te_ix,:])
        y_hat = pd.Series(y_hat, index=self.y.index)
        return Result(self.y, y_hat)

class Result(object):
    def __init__(self, y, y_hat):
        assert isinstance(y, pd.Series)
        assert isinstance(y_hat, pd.Series)
        # NOTE: y and y_hat are not assumed to be aligned, because
        #   a missing/NA value in y_hat means "no prediction"
        self.y = y.dropna()
        self.y_hat = y_hat.loc[self.y.index]

    @property
    def accuracy(self):
        return (self.y == self.y_hat).mean()

    @property
    def summary(self):
        return pd.Series([getattr(self, p) for p in self.PROPERTIES], 
                index=self.PROPERTIES)

class BinaryResult(Result):
    PROPERTIES = ["TP", "FP", "FN", "precision", "recall"]

    def __init__(self, y, y_hat):
        assert y.dtype.name == "bool"
        if y_hat.dtype.name == "float":
            raise NotImplementedError("Scores/probabilities not supported yet")
        assert y_hat.dtype.name == "bool"
        super(BinaryResult, self).__init__(y, y_hat)

    @property
    def TP(self):
        return (self.y & self.y_hat).sum()

    @property
    def FP(self):
        return ((~self.y) & self.y_hat).sum()

    @property
    def FN(self):
        return ((self.y) & (~self.y_hat)).sum()

    @property
    def precision(self):
        return self.TP / (self.TP + self.FP)

    @property
    def recall(self):
        return self.TP / (self.TP + self.FN)

    """
    @property
    def AUC(self):
        try:
            return sklearn.metrics.roc_auc_score(self.y, self.y_hat)
        except:
            return np.nan
    """

class CategoricalResult(Result):
    def __init__(self, y, y_hat):
        assert y.dtype.name == "category"
        assert y_hat.dtype.name == "category"

        y = y.cat.remove_unused_categories()
        super(CategoricalResult, self).__init__(y, y_hat)

        categories = self.y.cat.categories
        self.y = self.y.cat.set_categories(categories, ordered=True)

        disjoint = len(set(self.y_hat.cat.categories) - set(categories)) > 0
        if disjoint:
            categories = ["OTHER"] + list(sorted(categories))
            self.y = self.y.cat.set_categories(categories)

            self.y_hat = pd.Series(list(self.y_hat), index=self.y_hat.index)
            self.y_hat.ix[~self.y_hat.isin(self.y.cat.categories)] = "OTHER"
            self.y_hat = self.y_hat.astype("category")
        self.y_hat = self.y_hat.cat.set_categories(categories)

        self._results = {}
        for c in categories:
            self._results[c] = BinaryResult(self.y == c, self.y_hat == c)

    @property
    def categories(self):
        return list(self.y.cat.categories)

    def confusion_matrix(self, normalize=False):
        o = pd.DataFrame(index=self.categories, columns=self.categories).fillna(0)
        o.index.name = "y"
        o.columns.name = "y_hat"
        for c_y in self.categories:
            for c_y_hat in self.categories:
                o.loc[c_y,c_y_hat] = ((self.y == c_y) & (self.y_hat == c_y_hat)).sum()
        if "OTHER" in self.categories:
            o = o.drop("OTHER")
        if normalize:
            o = o.divide(1 + o.sum(axis=0), axis=1)
        return o

    @property
    def summary(self):
        ix = list(self._results.keys())
        o = pd.concat([self._results[c].summary for c in ix], axis=1).T
        o.index = ix
        return o.rename_axis("Category")

class ContinuousResult(Result):
    PROPERTIES = ["accuracy", "r2", "MAD", "MSE"]

    def __init__(self, y, y_hat, tolerance=0.1):
        """
        Arguments
        ---------

        tolerance : float
            y and y_hat values less than this distance away are considered to be
            the same for classification-based metrics.
        """
        super(ContinuousResult, self).__init__(y, y_hat)
        self.tolerance = tolerance

    @property
    def micro_precision(self):
        raise NotImplementedError

    @property
    def micro_recall(self):
        raise NotImplementedError

    @property
    def macro_precision(self):
        y, y_hat = align_series(self.y, self.y_hat, dropna=True)
        return np.isclose(y, y_hat, atol=self.tolerance).mean()

    @property
    def macro_recall(self):
        y_hat = self.y_hat.loc[self.y.index]
        return np.isclose(self.y, y_hat, atol=self.tolerance).mean()

    @property
    def r2(self):
        y, y_hat = align_series(self.y, self.y_hat, dropna=True)
        return scipy.stats.pearsonr(y, y_hat)[0] ** 2

    @property
    def MAD(self):
        return np.abs(self.y - self.y_hat).mean()

    @property
    def MSE(self):
        return ((self.y - self.y_hat) ** 2).mean()

    def plot(self, path=None):
        plt.clf()
        y, y_hat = align_series(self.y, self.y_hat, dropna=True)
        data = pd.DataFrame.from_dict({"y":y, "y_hat":y_hat})
        sns.regplot(x="y", y="y_hat", data=data, fit_reg=False)
        plt.xlim([0, y.max() * 1.1])
        plt.ylim([0, y_hat.max() * 1.1])
        plt.ylabel("$\hat{y}$")
        plt.xlabel("y")
        if path is not None:
            plt.savefig(path)



"""
Linear models.
"""

import scipy.stats
import pandas as pd
import numpy as np
import statsmodels.api as sm

def mfit(Y, design, formula=None):
    """
    Fit multiple linear models, one for each row of Y.
    """
    assert Y.isnull().sum().sum() == 0
    assert Y.shape[1] == design.shape[0]
    n = Y.shape[1]

    X = np.array(design)
    parameters = design.columns
    beta, _, rank, sv = np.linalg.lstsq(X, Y.T)
    #residual = Y.T - X @ beta
    n = X.shape[0]
    X_se = X.std(axis=0) / np.sqrt(n)
    coef = pd.DataFrame(beta.T, index=Y.index, columns=parameters)
    t = coef / X_se
    p = pd.DataFrame(scipy.stats.t.sf(t.abs(), df=n-2) * 2,
            index=t.index, columns=t.columns)
    ts = t.sort_values("Hours", ascending=False)
    print(ts.head())
    print(p.loc[ts.index,:].head())
    print(p.describe())
    #print(t.head())
    #print(np.divide(beta, X.std(axis=0)))
    #t = pd.Series(beta / X.std(axis=0), index=parameters)
    #print(t)

    """
    residual = y  - X @ beta
    #r2 = (1 - residual) / ((y - y.var())
    t = pd.Series(beta / X.std(axis=0), index=parameters)
    p = scipy.stats.t.cdf(t, df=t.shape[0] - 1)
    p = pd.Series(np.minimum(p, 1-p) * 2, index=parameters)
    coef = pd.Series(beta, index=parameters)
    """



"""
Network helper functions.
"""

import base64
import contextlib
import http.client
import os
import os.path
import pathlib
import urllib.request
import zlib

from .log import LOG

def _decompress_stream(input_stream, output_stream):
    """
    Decompress the contents of input_stream to output_stream (assumes GZip format).
    """
    if not isinstance(input_stream, http.client.HTTPResponse):
        assert "b" in input_stream.mode
    assert "b" in output_stream.mode

    BLOCK_SIZE = 4096
    dc = zlib.decompressobj(16+zlib.MAX_WBITS)
    while True:
        data = input_stream.read(BLOCK_SIZE)
        if not data:
            break
        output_stream.write(dc.decompress(data))

def download(url, decompress=False, expire=None):
    """
    Download a file over HTTP or FTP, cache it locally, and return a path to it.

    Arguments
    ---------
    url : str
        The URL to download and cache.
    decompress : bool
        Whether to decompress the file before saving it (assumes gzip format).
    expire : numeric or None
        Expire the cache and redownload the file after this many days, 
        or if this parameter is None, never expire.

    Returns
    -------
    A :class:`pathlib.Path` object representing the path to the downloaded file
    on the file system.
    """
    # TODO: add "expire" parameter
    cache = os.path.join(os.path.expanduser("~/.cache/wrenlab/download/"))
    os.makedirs(cache, exist_ok=True)

    target_ascii = url.lower()
    if decompress:
        target_ascii += "-dc"
    target = base64.b64encode(target_ascii.encode("utf-8")).decode("utf-8")
    path = os.path.join(cache, target)

    if not os.path.exists(path):
        LOG.info("Downloading URL: {}".format(url))
        try:
            if decompress:
                LOG.debug("Decompressing stream (decompress = True)")
                with open(path, "wb") as o:
                    with contextlib.closing(urllib.request.urlopen(url)) as h:
                        _decompress_stream(h, o)
            else:
                urllib.request.urlretrieve(url, path)
            LOG.debug("URL successfully saved to: {}".format(path))
        except:
            if os.path.exists(path):
                os.unlink(path)
            LOG.debug("Failed to download URL: {}".format(url))
            raise
    return pathlib.Path(path)



"""
Python wrapper for the Aspera client (used for fast downloads from NCBI
server).
"""

# FIXME: if the user hasn't manually used ascp before, download()
# will hang, asking the user to trust the remote host and add
# remote key to PuTTY cache

# TODO: download binary if necessary?
# TODO: allow option to AsperaClient.download() to background the download
#   into other process or thread
# TODO: if binary is not found, add instructions into error message
#   about how to get it
# TODO: allow multiple source files to be given to download() ?

import os
import sys
import subprocess as sp
import tempfile

def _find_binary():
    """
    Search PATH for ascp. Returns None if it could not be found.
    """
    search_folders = list(sys.path)
    search_folders.extend([
        os.path.expanduser("~/.aspera/connect/bin/"),
        "/usr/bin/",
        "/usr/local/bin/",
        "/sbin/"])
    for folder in search_folders:
        path = os.path.join(folder, "ascp")
        if os.path.isfile(path) and os.access(path, os.X_OK):
            return path

class Error(IOError):
    pass

class Client(object):
    # NOTE: Newer versions of ascp use asperaweb_id_dsa.openssh, 
    #   whereas older ones use aspweraweb_id_dsa.putty
    # The right way to handle this would be to figure out when it changed
    # and detect ascp version.
    def __init__(self, 
            user="anonftp", 
            host="ftp-private.ncbi.nlm.nih.gov", 
            limit=200, # in MB
            key_file="~/.aspera/connect/etc/asperaweb_id_dsa.openssh"):
        assert isinstance(limit, int)

        self._binary_path = _find_binary()
        if not self._binary_path:
            raise Error("Aspera client binary (ascp) could not be found.")

        self._user = user
        self._host = host
        self._limit = limit
        self._key_file = os.path.expanduser(key_file)

    def download(self, source, destination, decompress=None):
        """
        Download a single file from this Aspera server,
        returning an open file handle to the resulting file.

        This handle will delete the underlying temporary file when it is
        closed, so use shutil.copyfile or similar to persist the data.
        """
        assert decompress is None, "Decompression not implemented."

        source_uri = "%s@%s:/%s" % (self._user, self._host, source.lstrip("/"))
        #handle = tempfile.NamedTemporaryFile(delete=True)
        #destination = handle.name
        args = [self._binary_path, 
                "-T",
                "-i", self._key_file, 
                "-l", "%sM" % self._limit,
                source_uri, destination]
        p = sp.Popen(args, stdin=sp.PIPE, stdout=sp.DEVNULL, stderr=sp.PIPE)
        p.wait()
        if p.returncode:
            msg = p.stderr.read().decode("utf-8").strip()
            raise Error(msg)
        #return handle



__all__ = ["Experiment"]

import os
import os.path
import collections
import multiprocessing as mp

import sklearn.decomposition
import numpy as np
import scipy.stats
import patsy
import pandas as pd
import statsmodels.api as sm
import matplotlib.pyplot as plt
import seaborn as sns
from joblib import Parallel, delayed

import wrenlab.impute
import wrenlab.correlation
import wrenlab.normalize
from wrenlab.util import LOG

Model = collections.namedtuple("Model", "params,metadata")

"""
class Model(collections.OrderedDict):
    ""
    Represents a (per-gene) collection of OLS models performed 
    on the same dataset and formula.
    ""
    def __init__(self, formula=None):
        self.formula = formula
        super(Model, self).__init__()

    def __repr__(self):
        return "<Model with {} elements>".format(len(self))

    @property
    def params(self):
        return pd.DataFrame([v.params for v in self.values()], index=self.keys())

    @property
    def p(self):
        return pd.DataFrame([v.pvalues for v in self.values()], index=self.keys())

    @property
    def t(self):
        return pd.DataFrame([v.tvalues for v in self.values()], index=self.keys())
"""

OLS = collections.namedtuple("OLS", "params,tvalues,pvalues")

def _fit(y, design, groups=None, regularize=False, method="statsmodels", formula=None):
    design, y = design.dropna().align(y.dropna(), axis=0, join="inner")

    if method == "statsmodels":
        if groups is not None:
            m = sm.MixedLM(y, design, groups=groups)
            if regularize:
                return m.fit_regularized()
            else:
                return m.fit()
        else:
            X = np.array(design)
            parameters = design.columns
            beta, _, rank, sv = np.linalg.lstsq(X, y)
            residual = y  - X @ beta
            #r2 = (1 - residual) / ((y - y.var())
            t = pd.Series(beta / X.std(axis=0), index=parameters)
            p = scipy.stats.t.cdf(t, df=t.shape[0] - 1)
            p = pd.Series(np.minimum(p, 1-p) * 2, index=parameters)
            coef = pd.Series(beta, index=parameters)
            return OLS(coef, t, p)
    elif method == "lmer":
        import wrenlab.R
        assert formula is not None
        fit = wrenlab.R.lmer(formula, y, design)
        df = fit.fixed_effects
        return OLS(df["coef"], df["t"], df["p"])
    else:
        raise ValueError

def _fit_multiple(Y, design):
    LOG.info("Fitting linear models: {} samples x {} transcripts".format(*Y.T.shape))
    X = np.array(design)
    parameters = design.columns
    beta, _, rank, sv = np.linalg.lstsq(X, Y.T)
    for j in range(beta.shape[1]):
        beta_j = beta[:,j]
        t = pd.Series(beta_j / X.std(axis=0), index=parameters)
        p = scipy.stats.t.cdf(t, df=t.shape[0] - 1)
        p = pd.Series(np.minimum(p, 1-p) * 2, index=parameters)
        coef = pd.Series(beta_j, index=parameters)
        yield OLS(coef, t, p)


class Experiment(object):
    def __init__(self, design, data, metadata=None):
        assert isinstance(data, pd.DataFrame)
        assert isinstance(design, pd.DataFrame)
        assert design.shape[0] == data.shape[1]
        assert (design.index == data.columns).all()
        self._design = design
        self._data = data
        self.metadata = metadata or {}

    @property
    def design(self):
        return self._design.copy()

    @property
    def data(self):
        return self._data.copy()

    @property
    def shape(self):
        return self._data.shape

    def complete(self):
        return not self._data.isnull().any().any()

    def impute(self, method="linear_model"):
        fn = getattr(wrenlab.impute, method)
        data = fn(self.data).dropna()
        return Experiment(self.design, data, 
                metadata={k:v for k,v in self.metadata.items()})

    def normalize(self, method="quantile"):
        """
        Normalize the data in this Experiment with the given method, 
        returning a new Experiment object.
        """
        assert method in ("vsn", "quantile")
        data = self._data.dropna(thresh=int(0.2 * self._data.shape[0]), axis=1)
        if data.shape[1] == 0:
            raise Exception("No samples remaining after dropping those with >80% missing values.")
        design = self.design.loc[data.columns,:]
        if method == "quantile":
            data = data.dropna(thresh=int(0.5 * self._data.shape[1]))
            if data.shape[0] == 0:
                raise Exception("No probes remaining after dropping those with >50% missing values.")
            data = wrenlab.normalize.quantile(data)
        elif method == "vsn":
            data = wrenlab.normalize.vsn(data.dropna(how="any"))
        else:
            raise ValueError("Invalid normalization method: '{}'".format(method))
        return Experiment(design, data, metadata={k:v for k,v in self.metadata.items()})

    def pca(self, n_components=10, subset=None):
        data = self.data.dropna()
        if subset is not None:
            ix = list(set(subset) & set(data.index))
            data = data.loc[ix,:]
        model = sklearn.decomposition.PCA(n_components=n_components)
        components = pd.DataFrame(model.fit_transform(data.T).T,
                columns=data.columns,
                index=["C{}".format(i) for i in range(1, n_components+1)])

        metadata = {k:v for k,v in self.metadata.items()}
        metadata["Original Dimensions"] = data.shape[1]
        return Experiment(self.design, components, metadata=metadata)

    def fit(self, formula, groups=None, regularize=False, method="statsmodels"):
        """
        Fit OLS models for each gene/probe to the given formula.

        Arguments
        ---------
        formula : str
            The statistical formula to fit.
        groups : str
            If this is supplied, fit a mixed-effects model using
            the provided categorical variable as the groups.
        regularize : bool
            Whether to regularize, which penalizes fixed-effects terms. 
            Applies only to mixed-effects models.
        method : ("statsmodels", "lmer")
            The fitting software to use.
        """
        assert method in ("statsmodels", "lmer")

        design = self.design.dropna()
        design, data = design.T.align(self._data, join="inner", axis=1)
        design = design.T
        data = data.dropna(thresh=max(3, int(0.1 * data.shape[1])))

        design["Age"] = design["Age"].astype(float)
        for c in ["TissueID", "PlatformID", "ExperimentID"]:
            if c in design.columns:
                design[c] = design[c].astype("category")

        if method == "statsmodels":
            design = patsy.dmatrix(formula, design, return_type="dataframe")
            if groups is not None:
                groups = design[groups]

        """
        print(design["Age"].min(), design["Age"].max())
        print(data.min().min(), data.max().max())
        print(data.isnull().sum().sum())
        print(data.head().T.head().T)
        print(design.head())
        """

        index = data.index
        LOG.info("Fitting linear model ...")
        LOG.info("Design shape: {}".format(design.shape))
        LOG.info("Data shape: {}".format(data.shape))

        #NJOBS = 1
        NJOBS = mp.cpu_count() - 8

        if (groups is None) and (data.isnull().sum().sum() == 0):
            rs = _fit_multiple(data, design)
        else:
            rs = Parallel(n_jobs=NJOBS)(delayed(_fit)
                    (data.loc[probe_id,:], design, groups, regularize, method, formula)
                        for probe_id in index)
        o = []
        for model in rs:
            df = pd.concat([model.params, model.tvalues, model.pvalues], 
                    keys=["coef", "t", "p"],
                    axis=1)
            v = pd.Series(df.values.ravel("C"),
                    index=pd.MultiIndex.from_product([df.index, df.columns],
                        names=["Variable", "Statistic"]))
            o.append(v)
        params = pd.concat(o, keys=self._data.index, axis=1).T
        params.index.name = self._data.index.name
        for axis in (0,1):
            params.sort_index(axis=axis, inplace=True)
        metadata = {k:v for k,v in self.metadata.items()}
        return Model(params=params, metadata=metadata)
 
        """
        print("job's done")
        for probe_id, fit in zip(index, rs):
            model[probe_id] = fit
        #for probe_id in self._data.index:
        #    y = self._data.loc[probe_id,:]

        return model
        """

    @property
    def report(self):
        mu = self._data.mean(axis=1).dropna()
        std = self._data.std(axis=1).dropna()
        ix = list(set(mu.index) & set(std.index))
        r = scipy.stats.pearsonr(mu.loc[ix], std.loc[ix])[0]

        data = np.array(self._data).flat
        data = data[~np.isnan(data)]
        return pd.Series((r, data.min(), data.max(), scipy.stats.skew(data)),
                index=[
                    "Mean-SD correlation",
                    "min",
                    "max",
                    "skewness",
        ])

        """
        os.makedirs(outdir, exist_ok=True)

        # Mean-SD plot
        mu = self._data.mean(axis=1)
        std = self._data.std(axis=1)
        sns.regplot(mu, std)
        plt.xlabel("Mean")
        plt.ylabel("Standard Deviation")
        plt.savefig(os.path.join(outdir, "mean_stdev.png"))

        # Distribution
        plt.clf()
        xs = np.array(self._data).flat
        sns.distplot(xs[~np.isnan(xs)])
        plt.xlabel("Intensity")
        plt.ylabel("Frequency")
        plt.savefig(os.path.join(outdir, "intensity_histogram.png"))
        """

    def correlate(self, covariate, method="pearson"):
        if method == "pearson":
            return wrenlab.correlation.pearson(self.data, self.design[covariate]).dropna()
        elif method == "spearman":
            return wrenlab.correlation.spearman(self.data, self.design[covariate]).dropna()
        else:
            raise ValueError("Method must be one of: {}".format(("pearson", "spearman")))

class ExperimentSet(list):
    def __init__(self, items, metadata=None):
        super(ExperimentSet, self).__init__(items)
        self.metadata = metadata or {}

    def plot_covariate_distribution(self, covariate):
        """
        Make a multi-boxplot of the distribution of values for a
        given covariate.
        """
        data = pd.DataFrame.from_records([(E.metadata["ExperimentID"], y) 
            for E in self for y in E.design[covariate].dropna()],
            columns=["ExperimentID", covariate])
        return sns.boxplot(y="ExperimentID", x=covariate, data=data, orient="h")

    def combine(self):
        """
        Combine the experiments within this ExperimentSet into a single Experiment
        with an additional ExperimentID categorical variable added to the design 
        matrix.
        """
        data = pd.concat([E.data for E in self], axis=1)
        design = pd.concat([E.design for E in self])
        parameters = set([k for E in self for k in E.metadata.keys()])
        for p in parameters:
            design[p] = 0
            for E in self:
                design.loc[E.design.index, p] = int(E.metadata[p])
            design[p] = design[p].astype("category")
            if len(set(design[p])) == 1:
                del design[p]
        return Experiment(design, data, metadata={k:v for k,v in self.metadata.items()})

    def filter(self, fn):
        """
        Filter experiments using a function.

        Arguments
        ---------
        fn : A function of signature f(Experiment) -> bool, where the result
            specifies whether to keep a given experiment.

        Returns
        -------
        A new :class:`ExperimentSet` with the selected experiments retained.
        """
        return ExperimentSet(list(filter(fn, self)),
                metadata={k:v for k,v in self.metadata.items()})



import os
import os.path
import shelve

class KVStore(shelve.DbfilenameShelf):
    """
    NOTE: keys are coerced to strings, so iterating through keys() or items()
    will be messed up.

    Values can be arbitrary objects.
    """

    def __init__(self, name, writeback=False):
        path = KVStore._path(name)
        os.makedirs(os.path.dirname(path), exist_ok=True)
        flag = "c" if writeback else "r"
        super(KVStore, self).__init__(path, flag=flag, writeback=writeback)

    @staticmethod
    def _path(name):
        return os.path.expanduser("~/.cache/wrenlab/kvstore/{}.bdb".format(name))

    @staticmethod
    def exists(name):
        return os.path.exists(KVStore._path(name))

    def __getitem__(self, key):
        return super(KVStore, self).__getitem__(str(key))

    def __setitem__(self, key, value):
        super(KVStore, self).__setitem__(str(key), value)





"""
Algorithms for gene/transcript function prediction.

References
----------

Dozmorov, MG, Giles, CB, Wren, JD (2011). Predicting gene ontology from a
global meta-analysis of 1-color microarray experiments. BMC Bioinformatics, 12
Suppl 10:S14.

Wren, JD (2009). A global meta-analysis of microarray expression data to
predict unknown gene functions and estimate the literature-data divide.
Bioinformatics, 25, 13:1694-701.
"""



from os.path import dirname, join as pjoin, exists as pexists
from pip.req import parse_requirements
from pip.download import PipSession
from setuptools import setup, find_packages, Extension

VERSION = "0.1.1.post6"

DEPENDENCIES = [
    "cython==0.23",
    "numpy", 
    "scipy", 
    "pandas", 
    "rpy2", 
    "patsy",
    "statsmodels", 
    "joblib", 
    "networkx", 
    "seaborn",
    "pyyaml",
    "scikit-learn",

    "dask",
    "blaze",
    "cloudpickle"
]

#requirements = pjoin(dirname(__file__), "requirements.txt")
#assert pexists(requirements)
#requirements = [str(r.req) for r in parse_requirements(requirements, session=PipSession())]

cmdclass = {}

###################
# Cython extensions
###################

extensions = [
        Extension("wrenlab.correlation",
            sources=["wrenlab/correlation.pyx"],
            extra_compile_args=['-fopenmp'],
            extra_link_args=['-fopenmp'],
            language="c++"),
        Extension("wrenlab.text.ahocorasick.types",
            sources=["wrenlab/text/ahocorasick/types.pyx"],
            language="c++"),
        Extension("wrenlab.text.ahocorasick.algorithm",
            sources=["wrenlab/text/ahocorasick/algorithm.pyx"],
            language="c++")
]

try:
    from Cython.Distutils import build_ext
    cmdclass["build_ext"] = build_ext
except ImportError:
    # FIXME
    pass

include_dirs = []
try:
    import numpy
    include_dirs.append(numpy.get_include())
except ImportError:
    pass

####################
# Package definition
####################

setup(
    name="wrenlab",
    version=VERSION,
    description="Wren Lab bioinformatics utilities",
    author="Cory Giles",
    author_email="mail@corygil.es",
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: GNU Affero General Public License v3 or later (AGPLv3+)",
        "Natural Language :: English",
        "Operating System :: POSIX",
        "Programming Language :: Python :: 3.5",
        "Topic :: Scientific/Engineering :: Bio-Informatics"
    ],
    license="AGPLv3+",

    include_package_data=True,
    py_modules=["mmat", "acfsa"],
    packages=find_packages(),
    install_requires=DEPENDENCIES,
    ext_modules=extensions,
    include_dirs=include_dirs,
    cmdclass=cmdclass

    #py_modules=["mmat", "matrixdb"],
    #entry_points={
    #    "console_scripts": [
    #        "mmat = mmat:cli"
    #    ]
    #},
)



import functools
import os.path

import numpy as np
import pandas as pd

import mmat

from wrenlab.util import memoize
from wrenlab.experiment import Experiment, ExperimentSet

@functools.lru_cache()
@memoize
def array_describe(path):
    # FIXME: Find a way to regenerate if mtime changes
    X = mmat.MMAT(path)
    return X.describe()

class GEO(object):
    def __init__(self, matrix_dir, labels):
        self._matrix_dir = matrix_dir
        self._labels = labels

    def __getitem__(self, taxon_id):
        return GEOTaxon(taxon_id, self._matrix_dir, self._labels)

class GEOTaxon(object):
    def __init__(self, taxon_id, matrix_dir, labels):
        self.taxon_id = int(taxon_id)
        self._path = os.path.join(matrix_dir, "{}.mmat".format(taxon_id))
        self._X = mmat.MMAT(self._path)
        self._path_t = os.path.join(matrix_dir, "{}.t.mmat".format(taxon_id))
        self._XT = mmat.MMAT(self._path_t)

        self._A = labels.copy()
        # FIXME: Drop Channel index, if it exists - assumes Channel is always 1 for now
        if isinstance(self._A.index, pd.MultiIndex):
            assert (self._A.index.names == ["SampleID", "Channel"]).all()
            self._A.index = self._A.index.droplevel(1)
        if self._A.index.dtype == object:
            assert self._A.index[0].startswith("GSM")
            self._A.index = [int(x[3:]) for x in self._A.index]
        assert self._A.index.dtype == np.int64

        self._A = self._A.ix[self._A["TaxonID"] == taxon_id,:] # necessary?
        X_ix = [int(x[3:]) for x in self._X.index]
        self._A = self._A.loc[list(set(self._A.index) & set(X_ix)),:]

    def describe(self, axis=1):
        if axis == 1:
            return array_describe(self._path)
        elif axis == 0:
            return array_describe(self._path_t)
        else:
            raise ValueError

    def gene(self, gene_id, standardize=False):
        if standardize is True:
            design = self.gene(gene_id, standardize=False)
            y = design["Expression"]
            summary = self.describe().dropna()
            summary.index = [int(x[3:]) for x in summary.index]
            ix = list(set(y.index) & set(summary.index))
            y,summary = y.loc[ix],summary.loc[ix,:]
            design = design.loc[ix,:]
            y = pd.Series([(yy - mu) / sd for yy,mu,sd in 
                zip(y,summary["mean"].astype(float),summary["std"].astype(float))],
                index=ix)
            design["Expression"] = y
            return design
        else:
            ix = ["GSM{}".format(ix) for ix in self._A.index]
            x = self._XT.loc[gene_id, ix].dropna()
            x.index = [int(x[3:]) for x in x.index]
            o, x = self._A.copy().align(x, join="inner", axis=0)
            o["Expression"] = x
            return o

    def experiment(self, experiment_id):
        # FIXME: assumes A index is str
        raise NotImplementedError

        A = self._A.ix[self._A["ExperimentID"] == experiment_id,:]
        X = self._X.loc[list(set(A.index)),:].to_frame().T
        design = A.loc[X.columns,["TissueID", "Age", "Gender"]]
        metadata = {
            "ExperimentID": int(experiment_id),
            "TaxonID": self.taxon_id
        }
        return Experiment(design, X, metadata=metadata)

    def experiments(self, tissue_id, min_samples=10, min_transcripts=500):
        # FIXME: assumes A index is str
        raise NotImplementedError

        A = self._A.ix[self._A["TissueID"] == tissue_id,:]
        for (experiment_id, platform_id), df in A.groupby(["ExperimentID", "PlatformID"]):
            N = df.shape[0]
            if N < min_samples:
                continue
            X = self._X.loc[df.index,:].to_frame().T
            X.index.name = "Entrez Gene ID"
            if X.dropna(how="any").shape[0] < min_transcripts:
                continue
            design = df.loc[:,["Age", "Gender"]]
            E = Experiment(design, X, metadata={
                "ExperimentID": int(experiment_id),
                "TaxonID": self.taxon_id,
                "PlatformID": int(platform_id),
                "TissueID": tissue_id})
            yield E



"""
Pandas helper functions.
"""

import pandas as pd

def align_series(x, y, dropna=True):
    """
    Align two :class:`pandas.Series` using an inner join of their indices.

    Arguments
    =========
    x, y : :class:`pandas.Series`
    dropna : bool, optional
        Whether to drop missing elements from both :class:`pandas.Series`
        objects before performing the join.

    Returns
    =======
    A 2-tuple of :class:`pandas.Series`, the aligned versions of x and y.

    """
    assert isinstance(x, pd.Series)
    assert isinstance(y, pd.Series)
    if dropna is True:
        x = x.dropna()
        y = y.dropna()

    ix = list(set(x.index) & set(y.index))
    return x.loc[ix], y.loc[ix]

def sort_absolute_value(v, column=None, ascending=True):
    """
    Sort a :class:`pandas.Series` or a :class:`pandas.DataFrame` by
    absolute value (in the latter case, by the absolute value of a given column).

    Arguments
    =========
    v : :class:`pandas.Series` or :class:`pandas.DataFrame`
    column : str, optional
        If a :class:`pandas.DataFrame` is provided, sort by this column.
        Ignored if v is a :class:`pandas.Series`.
    ascending : bool
        Whether to sort ascending (default True).
    """
    assert isinstance(v, pd.Series) or isinstance(v, pd.DataFrame)

    if column is not None:
        df = v
        ix = sort_absolute_value(df[column], ascending=ascending).index
        return df.loc[ix,:]
    else:
        return v.loc[v.abs().sort_values(ascending=ascending).index]



"""
Plotting functions, mostly a wrapper layer to seaborn for easier usage.
"""

import collections
import functools

import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
import seaborn as sns

def plot_fn(fn):
    @functools.wraps(fn)
    def wrap(*args, x_lim=None, y_lim=None, y_label=None, x_label=None, title=None, **kwargs):
        plt.clf()
        figure = fn(*args, **kwargs)
        if x_label is not None:
            plt.xlabel(x_label)
        if y_label is not None:
            plt.ylabel(y_label)
        if title is not None:
            plt.title(title)
        return figure
    return wrap

"""
def rotate_xlabels(degrees=30):
    fig, ax = plt.subplots()
    labels = ax.get_xticklabels()
    plt.setp(labels, rotation=degrees)
"""

@plot_fn
def bar(xs, orient="v", error=None, **kwargs):
    assert isinstance(xs, pd.Series)
    if orient == "v":
        x,y = xs.index, xs
    else:
        x,y = xs, xs.index

    data = pd.DataFrame.from_dict({"x": x, "y": y})
    ax = sns.barplot(x="x", y="y", data=data, orient=orient, order=xs.index)
    if error is not None:
        if isinstance(error, collections.Iterable):
            assert len(error) == len(xs)
            error = list(error)
            for i in range(len(error)):
                ax.errorbar(i, xs.iloc[i], yerr=error[i], fmt="o", color="black")
        else:
            ax.errorbar(np.arange(len(xs)), xs, yerr=error, fmt="o", color="black")
    return plt.gcf()

def coef(fit):
    """
    Plot of statsmodels linear model coefficients.
    """
    return bar(fit.params, error=fit.bse)



import functools
import sqlite3
import urllib.request

import pandas as pd

import wrenlab.ncbi.taxonomy
from wrenlab.util import LOG, download

def _get_path():
    urls = [
        "http://gbnci.abcc.ncifcrf.gov/geo/GEOmetadb.sqlite.gz",
        "http://watson.nci.nih.gov/~zhujack/GEOmetadb.sqlite.gz",
        "http://dl.dropbox.com/u/51653511/GEOmetadb.sqlite.gz"
    ]
    for url in urls:
        try:
            return download(url, decompress=True, expire=30)
        except urllib.request.HTTPError:
            LOG.debug("HTTP 404: {}".format(url))
            continue
    raise Exception("Could not retrieve GEOmetadb SQLite database.")

def connect():
    path = str(_get_path())
    return GEOMetaDB(path)

class GEOMetaDB(sqlite3.Connection):
    def __init__(self, path):
        super(GEOMetaDB, self).__init__("file:///{}?mode=ro".format(path), uri=True)

        self._taxonomy = wrenlab.ncbi.taxonomy.names()
        self._taxon_name_id = dict(zip(self._taxonomy["Scientific Name"], 
            self._taxonomy.index))
        
    def __del__(self):
        self.close()

    @property
    @functools.lru_cache()
    def experiment_sample(self):
        q = """
        SELECT
            CAST(SUBSTR(gse,4) AS int) as 'ExperimentID',
            CAST(SUBSTR(gsm,4) AS int) as 'SampleID'
        FROM gse_gsm;
        """
        return pd.read_sql(q, self)

    @property
    @functools.lru_cache()
    def samples(self):
        o = []
        for i in range(1, 3):
            q = """
            SELECT 
                CAST(SUBSTR(gsm,4) AS int) AS 'SampleID', 
                {i} as Channel, 
                CAST(SUBSTR(gpl,4) AS int) AS 'PlatformID',
                organism_ch{i} as 'TaxonID',
                title as Title, 
                characteristics_ch{i} as Characteristics
            FROM gsm
            WHERE channel_count >= {i};
            """.format(i=i)
            o.append(pd.read_sql(q, self))
        o = pd.concat(o)

        ################################
        # TODO: handle multiple channels
        ################################
        o = o.query("Channel == 1").copy()#.drop("Channel", axis=1)
        o = o.set_index(["SampleID", "Channel"])

        o["TaxonID"] = [self._taxon_name_id.get(name) for name in o["TaxonID"]]
        o = o.dropna(subset=["TaxonID"])
        o["TaxonID"] = o["TaxonID"].astype(int)
        return o



"""
Distance metrics and nearest neighbors functions.
"""

import pandas as pd
import numpy as np

import wrenlab.correlation

import multiprocessing as mp
import multiprocessing.sharedctypes

class SharedArray(object):
    def __init__(self, X):
        self.size = X.size
        self.shape = X.shape
        self.buffer = multiprocessing.sharedctypes.RawArray("d", X)

    @property
    def data(self):
        X = np.ctypeslib.as_array(self.buffer)
        X.shape = self.shape
        return X

def _nearest_neighbors(X, i, k):
    r = wrenlab.correlation.pearson(X, X[i,:])["r"]
    d = np.array(1 - (r * 2))
    return np.argsort(d)[1:(k+1)]

from joblib import Parallel, delayed
from joblib.pool import has_shareable_memory

def nearest_neighbors(X, k=50):
    """
    Arguments
    ---------
    X : :class:`pandas.DataFrame`
        The expression matrix of dimensions (n,p), 
        with probes as rows and samples as columns.

    k : int, default 100
       The number of nearest neighbors. 

    Returns
    -------
    A :class:`pandas.DataFrame` of dimensions, (n,k) containing 
    the nearest neighbors.
    """
    assert k < (X.shape[0] - 1)
    assert X.T.isnull().sum().max() >= 3
    index = X.index
    X = np.array(X)
    N = np.zeros((X.shape[0], k), dtype=index.dtype)
    rs = Parallel(n_jobs=mp.cpu_count() - 2, max_nbytes=20e9)(
            delayed(_nearest_neighbors)(X, i, k) for i in range(X.shape[0]))
    for i,ix in enumerate(rs):
        N[i,:] = index[ix]
    return pd.DataFrame(N, index=index)



import collections
import os.path

import seaborn as sns
import matplotlib.pyplot as plt
import statsmodels.api as sm
import patsy
import pandas as pd
import numpy as np
from scipy import stats

import wrenlab.data.signature
import wrenlab.plot
import wrenlab.normalize
import wrenlab.ontology
import wrenlab.GO
from wrenlab.util import memoize, download, remote_table

@memoize
def _gene_ontology(taxon_id):
    o = wrenlab.ontology.fetch("GO")
    X = remote_table("NCBI.gene.gene_go")
    ix = (X["Taxon ID"] == taxon_id) & (X["Evidence"] != "IEA")
    mapping = X.ix[ix,["Gene ID", "Term ID"]]
    return AnnotationSet(mapping, metadata=o.terms.drop("Namespace", axis=1))

def _gene_ontology_concepts(taxon_id):
    A = _gene_ontology(taxon_id)
    J = A.mapping.drop(["Score"], axis=1)\
            .merge(wrenlab.GO.concepts())\
            .drop(["Term ID"], axis=1)\
            .loc[:,["Element ID", "Concept", "Direction"]]
    return AnnotationSet(J)

def _NoaLTP(taxon_id):
    assert taxon_id == 10090
    return wrenlab.data.signature.NoaLTP()

_AVAILABLE = {
    "GO": _gene_ontology,
    "gene_ontology": _gene_ontology,
    "gene_ontology_concepts": _gene_ontology_concepts,
    "NoaLTP": _NoaLTP
}

class SignatureAnalysis(object):
    def __init__(self, data):
        self._data = data

    @property
    def data(self):
        return self._data.copy()

    def plot(self, term, path, center=False):
        data = self._data.loc[term,:]
        xs = data["coef"]
        error = next(iter(data["stderr"]))
        if center:
            xs -= xs.iloc[0]
        wrenlab.plot.bar(xs, 
                error=error, 
                y_label="\n".join([term, r"(log$_2$ expression, relative LM coefficient)"]))
        plt.savefig(path)

    def plot_all(self, dir, center=False):
        assert os.path.isdir(dir)
        for term in self._data.index.levels[0]:
            path = os.path.join(dir, "{}.png".format(term.replace(" ", "_")))
            self.plot(term, path, center=center)

class AnnotationSet(object):
    """
    A bidirectional map of Entrez Gene IDs to Term IDs, with an additional
    "background" attribute listing all "elements" (usually genes)
    annotated by any term in the set.
    """

    def __init__(self, mapping, metadata=None):
        """
        Arguments
        ---------
        mapping: a :class:`pandas.DataFrame` with 2 or 3 columns: 
            Element IDs, Term IDs, and (optional) a score associated with the relationship
        metadata: :class:`pandas.DataFrame`, optional
            Term IDs are row indices, columns contain metadata about each term (e.g., names)
        """

        assert isinstance(mapping, pd.DataFrame)
        assert(mapping.shape[1] in (2,3))
        if mapping.shape[1] == 2:
            mapping["Score"] = 1
        mapping.columns = ["Element ID", "Term ID", "Score"]
        self.mapping = mapping.drop_duplicates(subset=["Element ID", "Term ID"]) 
        self.metadata = metadata

    def __repr__(self):
        return "<AnnotationSet with {} terms and {} elements>".format(len(self.terms),
                len(self.elements))

    @property
    def elements(self):
        return list(sorted(set(self.mapping["Element ID"])))

    @property
    def terms(self):
        return list(sorted(set(self.mapping["Term ID"])))

    def enrichment(self, scores, min_count=3):
        """
        Perform enrichment analysis for the association of a scored set of genes 
        with terms using Fisher's Exact test.

        Arguments
        ---------
        scores: a :class:`pandas.Series` with Entrez Gene IDs as indices and
            numeric or boolean values  
        min_count: int
            The minimum number of selected genes which must be annotated with the 
            term in order to consider the term for enrichment.

        Returns
        -------
        A :class:`pandas.DataFrame` with statistics showing enrichment values 
        for each term.
        """
        return self._enrichment_binary(scores, min_count)

    def _enrichment_permutation(self, scores):
        raise NotImplementedError

    def _enrichment_binary(self, scores, min_count):
        background = list(set(self.elements) & set(scores.index))
        scores = scores.loc[background]

        if (scores.dtype == np.bool):
            s = scores
        else:
            s = scores > scores.quantile(0.95)
        s = s.astype(int)
        s.name = "Score2"
        m = self.mapping.ix[self.mapping["Element ID"].isin(background),:].copy()
        m["Score"] = 1
        j = m.merge(s.to_frame(), how="inner", left_on="Element ID", right_index=True)\
                .dropna(subset=["Term ID"])
        counts = j.groupby(["Score2", "Term ID"])["Score"].count()

        # as a s
        ct = pd.concat([
                counts.loc[1],
                s.sum() - counts.loc[1],
                counts.loc[0]], axis=1)
        ct.columns = ["AS", "S", "A"]
        ct = ct.fillna(0)
        ct["O"] = len(s) - ct.sum(axis=1)
        ct = ct.ix[ct["AS"] >= min_count,:]
        assert ((ct["AS"] + ct["S"]) == s.sum()).all()
        assert ((ct["A"] + ct["O"]) == (len(s) - s.sum())).all()

        o = []
        for term_id in ct.index:
            cts = ct.loc[term_id,:]
            cts = np.array(cts).reshape((2,2))
            OR, p = stats.fisher_exact(cts)
            slpv = -1 * np.sign(np.log(OR)) * np.log10(p)
            o.append((term_id, OR, slpv))
        o = pd.DataFrame(o, columns=["Term ID", "Odds Ratio", "SLPV"])\
                .set_index(["Term ID"])
        o = ct.join(o, how="inner")
        if self.metadata is not None:
            o = self.metadata.join(o, how="right")
        return o.sort_values("SLPV", ascending=False)

    def signature(self, X, D, groups, subset=None):
        assert groups.shape[0] == 2 ** groups.shape[1]
        groups = groups.sort_values(by=list(groups.columns))
        n_coef = groups.shape[0]

        X = X.loc[list(set(X.index) & set(self.mapping["Element ID"])),:]
        Am = self.mapping.ix[self.mapping["Element ID"].isin(X.index),:]
        Xn = wrenlab.normalize.quantile(X)

        dvars = ":".join("C({})".format(c) for c in D.columns)
        formula = "Expression ~ C(Gene) + {} + Signature:{}".format(dvars, dvars)
        n_coef = groups.shape[0]

        o, ix = [], []

        for dcat, df in Am.groupby("Term ID"):
            s = np.array(df["Score"])
            if len(set(s)) < 2:
                continue
            n = df.shape[0]
            design = pd.concat([D]*n)
            design["Signature"] = np.tile(s, (D.shape[0],1)).T.flatten()
            design["Signature"] -= design["Signature"].mean()
            design["Gene"] = np.tile(df["Element ID"], (D.shape[0],1)).T.flatten()
            design["Expression"] = np.array(Xn.loc[df["Element ID"],:]).flatten()
            
            y, x = patsy.dmatrices(formula, design)
            model = sm.OLS(y, x)
            fit = model.fit()
            coef = fit.params
            stderr = fit.bse

            for g, coef, stderr in zip(groups.index, coef[-n_coef:], stderr[-n_coef:]):
                ix.append((dcat, g))
                o.append((n, coef, stderr))
        ix = pd.MultiIndex.from_tuples(ix, names=["Term ID", "Group"])
        o = pd.DataFrame.from_records(o, index=ix, columns=["N", "coef", "stderr"])
        return SignatureAnalysis(o)

    @staticmethod
    def get(key, taxon_id):
        return annotation(key, taxon_id)

def annotation(key, taxon_id):
    fn = _AVAILABLE[key]
    return fn(taxon_id)



"""
Signature analysis:

Problem
=======

Given a set of genes with numeric attributes (expression levels) and a
"database" of genes with numeric attributes, signature analysis provides
different methods of assessing the correlation between the two.

A concrete example: given a set of gene fold changes and a database of genes
that are positive and negative regulators of a certain pathway, is this pathway likely
to be up or down-regulated in this experiment?
"""


